import multiprocessing
import random


class ProxyGuard(object):
    """
    I'm a thread-safe guard.
    I have proxies, proxies have a number of threads that can use them at once.
    Someone can ask me to give him a proxy. I give it. If no proxies are available, I return None
    """

    def __init__(self, proxy_config):
        # {proxy: max_usages_at_once}
        self.counter = proxy_config
        self.lock = multiprocessing.Lock()

    def acquire_proxy(self):
        self.lock.acquire(True)

        selected_proxy = None
        items = list(self.counter.items())
        random.shuffle(items)
        for key, val in items:
            if val > 0:
                selected_proxy = key
                self.counter[key] -= 1
                break

        self.lock.release()

        return selected_proxy

    def release_proxy(self, proxy):
        self.lock.acquire(True)
        self.counter[proxy] += 1
        self.lock.release()



from utils_ak.portion import *
from utils_ak.coder import cast_js
from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock
from utils_ak.block_tree.block import Block

from loguru import logger


def validate_disjoint_by_axis(b1, b2, axis=0):
    validate_disjoint_by_intervals((b1.x[axis], b1.y[axis]), (b2.x[axis], b2.y[axis]))


def validate_disjoint_by_intervals(i1, i2):
    try:
        disposition = int(i1[1] - i2[0])
    except:
        disposition = 1

    i1 = cast_interval(*i1)
    i2 = cast_interval(*i2)

    assert calc_interval_length(i1 & i2) == 0, cast_js({"disposition": disposition})


def test_validate_disjoint_by_axis():
    print("Validate disjoint test")
    for t in range(0, 10):
        a = ParallelepipedBlock("a", n_dims=1, x=[t], size=[4])
        b = ParallelepipedBlock("b", n_dims=1, x=[3], size=[3])
        print(a, b)
        try:
            validate_disjoint_by_axis(a, b, 0)
        except AssertionError as e:
            print("AssertionError on disposition", e)


class ClassValidator:
    def __init__(self, window=2, window_by_classes=None):
        self.validators = {}
        self.window = window
        self.window_by_classes = window_by_classes or {}

    def add(self, class1, class2, validator, uni_direction=False):
        self.validators[(class1, class2)] = validator
        if uni_direction:
            self.validators[(class2, class1)] = validator

    def validate(self, b1, b2):
        key = (b1.props["cls"], b2.props["cls"])
        if key in self.validators:
            self.validators[key](b1, b2)

    def __call__(self, parent, block):
        parent_blocks = parent.children[-self.window :]

        if not parent_blocks:
            return

        b2 = block

        if not self.window_by_classes.get(block.props["cls"]):
            for b1 in parent_blocks:
                self.validate(b1, b2)
        else:
            parent_classes = set([b.props["cls"] for b in parent_blocks])
            for parent_cls in parent_classes:
                if parent_cls not in self.window_by_classes[block.props["cls"]]:
                    # don't check at all
                    continue
                cls_parent_blocks = [
                    b for b in parent_blocks if b.props["cls"] == parent_cls
                ]
                cls_parent_blocks = cls_parent_blocks[
                    -self.window_by_classes[block.props["cls"]][parent_cls] :
                ]

                for b1 in cls_parent_blocks:
                    self.validate(b1, b2)


def test_class_validator():
    class_validator = ClassValidator(window=1)
    class_validator.add("a", "a", validator=validate_disjoint_by_axis)

    root = ParallelepipedBlock("root", axis=0)
    a1 = ParallelepipedBlock("a", n_dims=2, x=[0, 0], size=[5, 1])
    a2 = ParallelepipedBlock("a", n_dims=2, x=[2, 0], size=[5, 1])
    b = ParallelepipedBlock("b", n_dims=2, x=[0, 1], size=[5, 1])
    root.add_child(a1)

    try:
        class_validator(root, a2)
    except AssertionError as e:
        print(e)

    class_validator(root, b)

    root.add_child(b)

    # window is 1 - validation should pass now
    class_validator(root, a2)

    root.add_child(a2)


if __name__ == "__main__":
    test_validate_disjoint_by_axis()
    test_class_validator()



from dynaconf import Dynaconf, Validator

settings = Dynaconf(
    settings_files=[
        "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/2020.09-qset/qset-feature-store/qset_feature_store/configs/default_settings.yaml",
        "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/2020.09-qset/qset-feature-store/qset_feature_store/configs/settings.yaml",
        "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/2020.09-qset/qset-feature-store/qset_feature_store/configs/settings.local.yaml",
        "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/2020.09-qset/qset-feature-store/qset_feature_store/configs/worker_settings.yaml",
        "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/2020.09-qset/qset-feature-store/qset_feature_store/configs/.secrets.yaml",
    ],
    environments=True,
    load_dotenv=True,
    # load_dotenv={"when": {"env": {"is_in": ["development"]}}},
    envvar_prefix="DYNACONF",
    env_switcher="ENV_FOR_DYNACONF",
    dotenv_path="configs/.env",
)


def test():
    from pprint import pprint

    with settings.using_env("default"):
        pprint(settings.as_dict())
    pprint(settings.as_dict())


if __name__ == "__main__":
    test()



import os
import glob

from utils_ak.os import list_files


class SplitFile:
    def __init__(self, fn):
        self.fn = fn

    def list(self):
        return list_files(
            os.path.dirname(self.fn),
            pattern=os.path.join(
                os.path.dirname(self.fn), os.path.splitext(os.path.basename(self.fn))[0]
            )
            + "*",
        )

    def get_indexes(self):
        cur_indexes = []
        for fn in self.list():
            if os.path.abspath(fn) == os.path.abspath(self.fn):
                cur_indexes.append(0)
            else:
                cur_indexes.append(int(os.path.splitext(fn)[0].rsplit("_", 1)[-1]))
        return cur_indexes

    def get_current(self, max_size=None):
        cur_indexes = self.get_indexes()

        if not cur_indexes:
            return

        suffix = "" if max(cur_indexes) == 0 else "_{}".format(max(cur_indexes))
        cur_fn = os.path.splitext(self.fn)[0] + suffix + os.path.splitext(self.fn)[-1]

        if max_size and os.path.getsize(cur_fn) > max_size:
            return self.get_new()
        else:
            return cur_fn

    def get_new(self):
        cur_indexes = self.get_indexes()
        if not cur_indexes:
            return self.fn

        suffix = "_{}".format(max(cur_indexes) + 1)
        return os.path.splitext(self.fn)[0] + suffix + os.path.splitext(self.fn)[-1]


def test_split_file():
    sf = SplitFile("test.csv")
    print(sf.list(), sf.get_indexes(), sf.get_current(), sf.get_new())
    with open(sf.get_new(), "w") as f:
        pass
    print(sf.list(), sf.get_indexes(), sf.get_current(), sf.get_new())
    with open(sf.get_new(), "w") as f:
        pass
    print(sf.list(), sf.get_indexes(), sf.get_current(), sf.get_new())

    for fn in sf.list():
        print("Removing", fn)
        os.remove(fn)


if __name__ == "__main__":
    test_split_file()



import time
import asyncio

from utils_ak.zmq import endpoint
from utils_ak.simple_microservice import SimpleMicroservice
from utils_ak.loguru import configure_loguru_stdout


class Ping(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__("Ping", *args, **kwargs)
        self.add_callback("ping", "", self.send_ping)
        self.add_timer(self.send_ping, interval=1, n_times=1, args=("init", "init"))

    def send_ping(self, topic, msg):
        self.logger.info(f"Received", topic=topic, msg=msg)
        time.sleep(1)
        self.publish("pong", "", msg="ping")


class Pong(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__("Pong", *args, **kwargs)
        self.add_callback("pong", "", self.send_pong)

    def send_pong(self, topic, msg):
        self.logger.info(f"Received", topic=topic, msg=msg)
        time.sleep(1)
        self.publish("ping", "", msg="pong")


def run_ping():
    configure_loguru_stdout("TRACE")
    ping = Ping(
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "ping": {"type": "sub", "endpoint": endpoint("localhost", 6554)},
                    "pong": {"type": "sub", "endpoint": endpoint("localhost", 6555)},
                }
            },
        ),
    )

    ping.run()


def run_pong():
    configure_loguru_stdout("TRACE")

    Pong(
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "ping": {"type": "sub", "endpoint": endpoint("localhost", 6554)},
                    "pong": {"type": "sub", "endpoint": endpoint("localhost", 6555)},
                }
            },
        ),
    ).run()


if __name__ == "__main__":
    import multiprocessing

    multiprocessing.Process(target=run_ping).start()
    multiprocessing.Process(target=run_pong).start()



import copy

from utils_ak.numeric import *
from utils_ak.simple_vector import *


from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock
from utils_ak.block_tree.block import Block
from utils_ak.block_tree.validation import validate_disjoint_by_axis
from utils_ak.block_tree.pushers.pushers import simple_push

from loguru import logger


def dummy_push(
    parent, block, validator, max_tries=24, start_from="last_end", iter_props=None
):
    axis = parent.props["axis"]

    if is_int_like(start_from):
        cur_start = int(float(start_from))
    elif start_from == "beg":
        cur_start = block.props["x_rel"][axis]
    else:
        if not parent.children:
            cur_start = 0
        else:
            if start_from == "last_beg":
                cur_start = max(
                    [child.props["x_rel"][axis] for child in parent.children]
                )
            elif start_from == "last_end":
                cur_start = max(
                    [
                        (child.props["x_rel"] + child.size)[axis]
                        for child in parent.children
                    ]
                )
            else:
                raise Exception("Unknown beg type")

    iter_props = iter_props or [{}]

    cur_x = cast_simple_vector(block.n_dims)

    cur_try = 0

    # logger.debug('Inserting blocks', parent=parent.props['cls'], block=block.props['cls'])

    while cur_try < max_tries:
        dispositions = []
        for props in iter_props:
            # try to push
            props = copy.deepcopy(props)
            cur_x[axis] = cur_start
            props["x"] = cur_x
            # logger.debug('Trying to push from ', cur_start=cur_start)
            res = simple_push(parent, block, validator=validator, new_props=props)

            if isinstance(res, Block):
                # success
                # logger.debug('Success')
                return block
            else:
                assert isinstance(res, dict)

                if "disposition" in res:
                    dispositions.append(res["disposition"])

        disposition = min(dispositions) if len(dispositions) == len(iter_props) else 1
        # logger.debug('Disposition', disposition=disposition)
        cur_start += disposition
    raise Exception("Failed to push element")


def test_dummy_push():
    from utils_ak.loguru import configure_loguru_stdout

    configure_loguru_stdout()

    def brute_validator(parent, block):
        for c in parent.children:
            validate_disjoint_by_axis(c, block, axis=parent.props["axis"])

    logger.debug("Dummy push test")
    root = ParallelepipedBlock("root", n_dims=1, x=[2], axis=0)
    a = ParallelepipedBlock("a", n_dims=1, size=[4], axis=0)
    b = ParallelepipedBlock("b", n_dims=1, size=[3], axis=0)
    dummy_push(root, a, brute_validator)
    dummy_push(root, b, brute_validator, start_from=0)
    logger.debug("Root", root=root)


if __name__ == "__main__":
    test_dummy_push()



from utils_ak.job_orchestrator.worker.test.test_worker import *
from utils_ak.job_orchestrator.config import settings
from utils_ak.deployment import *

MESSAGE_BROKER = settings.as_dict()["TRANSPORT"]["message_broker"]


def test_batch():
    test_microservice_worker(
        TestWorker,
        {"type": "batch", "message_broker": MESSAGE_BROKER},
        run_listener=True,
    )


def test_streaming():
    test_microservice_worker(
        TestWorker,
        {"type": "streaming", "message_broker": MESSAGE_BROKER},
        run_listener=True,
    )


def test_deployment():
    controller = ProcessController()
    test_microservice_worker_deployment(
        "../worker/test/sample_deployment.yml", controller, MESSAGE_BROKER
    )


if __name__ == "__main__":
    test_batch()
    # test_streaming()
    # test_deployment()



from clickhouse_driver import Client


def _create_clickhouse_table(cli, df, db, table, partition_by, order_by):
    types_map = {
        "int64": "UInt64",
        "float64": "Float64",
        "object": "String",
        "datetime64[ns]": "DateTime",
    }

    create_table_query = """CREATE TABLE {db}.{table} (
        {dtypes}
    ) ENGINE = MergeTree
    PARTITION BY {partition_by}
    ORDER BY {order_by}
    SETTINGS index_granularity = 8192;"""

    dtypes = ",\n".join(
        [
            "{} {}".format(df.columns[i], types_map[str(dtype)])
            for i, dtype in enumerate(df.dtypes)
        ]
    )

    query = create_table_query.format(
        table=table,
        db=db,
        dtypes=dtypes,
        partition_by=partition_by,
        order_by=order_by,
    )
    return cli.execute(query)


def _insert_df(cli, df, db, table):
    cli.execute(f"INSERT INTO {db}.{table} VALUES", df.to_dict("records"))


def ingest_table(df, host, db, table, partition_by, order_by):
    cli = Client(host=host)
    _create_clickhouse_table(cli, df, db, table, partition_by, order_by)
    _insert_df(cli, df, db, table)


def test():
    from utils_ak.clickhouse.config import load_config
    from datetime import datetime

    CONFIG = load_config("test")
    print(CONFIG)
    import pandas as pd

    df = pd.DataFrame([[1, 1.0, "a", datetime.now()]], columns=["a", "b", "c", "d"])

    table = "test13"
    ingest_table(
        df, CONFIG["clickhouse_url"], "datasets", table, df.columns[0], df.columns[0]
    )

    from pandahouse import read_clickhouse

    connection = {
        "host": f'http://{CONFIG["clickhouse_url"]}:8123',
        "database": "datasets",
    }
    df = read_clickhouse(
        f"SELECT * FROM datasets.{table}", index_col=None, connection=connection
    )
    print(df)


if __name__ == "__main__":
    test()



from utils_ak.fluid_flow import *
import warnings

warnings.filterwarnings("ignore")


def test_flow_processor_1():
    container = Container("Input", value=100, max_pressures=[None, 10])

    processor = Processor("Output")

    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_2():
    container = Container("Input", value=100)
    processor = Processor(
        "Output", processing_time=5, max_pressures=[10, None], transformation_factor=2.0
    )
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_zero_pressure():
    container = Container("Input", value=100)
    processor = Processor("Output", max_pressures=[0, 0])
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_limit():
    container = Container("Input", value=100, max_pressures=[None, 10])
    processor = Processor("Output", limits=[50, None])
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


if __name__ == "__main__":
    test_flow_processor_1()
    test_flow_processor_2()
    test_flow_processor_zero_pressure()
    test_flow_processor_limit()



import time
import os
import asyncio
from utils_ak.simple_microservice import SimpleMicroservice
from utils_ak.coder import cast_dict_or_list


class Worker:
    def __init__(self, id, payload):
        self.id = id
        self.payload = payload

    def run(self):
        raise NotImplementedError


class MicroserviceWorker(Worker):
    def __init__(self, id, payload):
        super().__init__(id, payload)
        self.microservice = SimpleMicroservice(
            id, message_broker=self.payload["message_broker"]
        )

        self.microservice.add_timer(
            self.microservice.publish,
            interval=3.0,
            args=(
                "monitor_in",
                "heartbeat",
            ),
            kwargs={"id": self.id},
        )

        self.microservice.add_timer(
            self.process, n_times=1, counter_type="left"
        )  # run once on init

    def send_state(self, status, state):
        self.microservice.logger.debug("Sending state", status=status, state=state)
        self.microservice.publish(
            "monitor_in", "state", id=self.id, status=status, state=state
        )

    async def process(self):
        raise NotImplementedError

    def run(self):
        self.microservice.run()


def run_worker(worker_cls, worker_config):
    worker = worker_cls(worker_config["worker_id"], worker_config["payload"])
    worker.run()



import time
import multiprocessing


from utils_ak.loguru import configure_loguru_stdout
from utils_ak.simple_microservice import run_listener_async
from utils_ak.job_orchestrator.worker.test.test_worker import TestWorker
from utils_ak.job_orchestrator.monitor import Monitor

from utils_ak.job_orchestrator.config import settings

MESSAGE_BROKER = settings.as_dict()["TRANSPORT"]["message_broker"]


def run_monitor():
    configure_loguru_stdout("DEBUG")
    monitor = Monitor(MESSAGE_BROKER)
    monitor.microservice.run()


def run_worker():
    configure_loguru_stdout("DEBUG")
    worker = TestWorker(
        "WorkerId",
        {"type": "batch", "message_broker": MESSAGE_BROKER},
    )
    worker.run()


def test():
    configure_loguru_stdout("DEBUG")
    run_listener_async("monitor_out", message_broker=MESSAGE_BROKER)
    time.sleep(1)

    multiprocessing.Process(target=run_monitor).start()
    time.sleep(3)
    multiprocessing.Process(target=run_worker).start()


if __name__ == "__main__":
    test()



import pandas as pd


def cast_prop_values(parent, child, key):
    pv = None if not parent else parent[key]
    v = child.relative_props.get(key)
    return pv, v


def relative_acc(parent, child, key, default=None, formatter=None):
    if key in child.relative_props:
        res = child.relative_props[key]
    else:
        if callable(default):
            default = default()
        res = default

    if formatter:
        res = formatter(res)
    return res


def cumsum_acc(parent, child, key, default=None, formatter=None):
    if parent:
        return parent[key] + relative_acc(
            parent, child, key, default=default, formatter=formatter
        )
    return relative_acc(parent, child, key, default=default, formatter=formatter)


class DynamicProps:
    def __init__(
        self, props=None, formatters=None, accumulators=None, required_keys=None
    ):
        self.accumulators = accumulators or {}
        self.required_keys = required_keys or []
        self.formatters = formatters or {}

        self.relative_props = self._format_props(props or {})

        self.parent = None
        self.children = []

    def _format_props(self, props):
        props = dict(props)
        for k, fmt in self.formatters.items():
            if k in props:
                props[k] = fmt(k, props[k])
        return props

    @staticmethod
    def default_accumulator(parent, child, key):
        pv, v = cast_prop_values(parent, child, key)
        return v if v is not None else pv

    def update(self, **props):
        props = self._format_props(props)
        self.relative_props.update(**props)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def __getitem__(self, item):
        accumulator = self.accumulators.get(item, self.default_accumulator)
        return accumulator(self.parent, self, item)

    def get(self, item, default=None):
        res = self[item]
        if res is None:
            res = default
        return res

    def keys(self):
        parent_keys = [] if not self.parent else self.parent.keys()
        res = (
            parent_keys
            + list(self.accumulators.keys())
            + list(self.relative_props.keys())
            + self.required_keys
        )
        return list(set(res))

    def all(self):
        return {key: self[key] for key in self.keys()}


def test_dynamic_props():
    ACCUMULATORS = {"foo": cumsum_acc, "bar": relative_acc}

    def gen_props(props=None):
        return DynamicProps(
            props=props,
            accumulators=ACCUMULATORS,
            required_keys=["foo", "bar", "other"],
        )

    root = gen_props({"foo": 1, "bar": 5, "other": 1})
    child1 = gen_props({"foo": 2})
    child2 = gen_props({"foo": 3})
    root.add_child(child1)
    child1.add_child(child2)

    values = []
    for i, node in enumerate([root, child1, child2]):
        values.append(
            [node[key] for key in ["foo", "bar", "other", "non-existent_key"]]
        )
    print(pd.DataFrame(values, columns=["foo", "bar", "other", "non-existent_key"]))

    for node in [root, child1, child2]:
        print(node.keys(), node.all())


if __name__ == "__main__":
    test_dynamic_props()



STUB = "__stub"


class SimpleIterator:
    def __init__(self, lst, start_from=0):
        self.lst = lst
        self.current_index = start_from

    def current(self):
        return self.lst[self.current_index]

    def __len__(self):
        return len(self.lst)

    def forward(
        self, step=1, return_last_if_out=False, update_index=True, out_value=None
    ):
        if self.current_index >= len(self.lst) - step:
            if return_last_if_out:
                res = self.lst[-1]
            else:
                res = out_value
        else:
            res = self.lst[self.current_index + step]

        if update_index:
            self.current_index = min(self.current_index + step, len(self.lst) - 1)
        return res

    def next(self, return_last_if_out=False, update_index=True):
        return self.forward(1, return_last_if_out, update_index)

    def backward(
        self, step=1, return_first_if_out=False, update_index=True, out_value=None
    ):
        if self.current_index <= step - 1:
            if return_first_if_out:
                res = self.lst[0]
            else:
                res = out_value
        else:
            res = self.lst[self.current_index - 1]

        if update_index:
            self.current_index = max(self.current_index - step, 0)
        return res

    def prev(self, return_first_if_out=False, update_index=True):
        return self.backward(1, return_first_if_out, update_index)

    def iter(self, direction="up", step=1, limit=None):
        yield self.current()
        counter = 1
        while True:
            if counter == limit:
                break

            if direction == "up":
                next = self.forward(step, out_value=STUB)
            elif direction == "down":
                next = self.backward(step, out_value=STUB)

            if next == STUB:
                break

            yield next
            counter += 1

    def iter_sequences(self, n=2, method="all", none_object=None):
        assert method in ["all", "any", "any_prefix", "any_suffix"]

        if method == "all":
            for i in range(len(self.lst) - n + 1):
                yield self.lst[i : i + n]
        elif method.startswith("any"):
            if not self.lst:
                return
            # todo: refactor using compound lists
            nones = [none_object] * (n - 1)
            for i in range(len(self.lst) + n - 1):
                if method == "any_suffix" and i < n - 1:
                    continue

                if method == "any_prefix" and i > len(self.lst) - 1:
                    continue

                yield nones[i:] + self.lst[max(i + 1 - n, 0) : i + 1] + nones[
                    len(self.lst) + n - 2 - i :
                ]

    def reset(self):
        self.current_index = 0


def iter_sequences(lst, *args, **kwargs):
    yield from SimpleIterator(lst).iter_sequences(*args, **kwargs)


def iter_pairs(lst, *args, **kwargs):
    yield from iter_sequences(lst, 2, *args, **kwargs)


def test_simple_bounded_iterator():
    lst = [1, 2, 3, 4]
    it = SimpleIterator(lst)

    print("Iter up")
    for v in it.iter("up"):
        print(v)
    print("Iter down")
    for v in it.iter("down"):
        print(v)
    print("return_last_if_out: false")
    for i in range(5):
        print(it.next(return_last_if_out=False))
    print("return_last_if_out: true")
    for i in range(5):
        print(it.prev(return_first_if_out=True))

    print("Sequences")
    print("all-2")
    for seq in it.iter_sequences(2):
        print(seq)

    print("any-2")
    for seq in it.iter_sequences(2, method="any"):
        print(seq)

    print("all-5")
    for seq in it.iter_sequences(5):
        print(seq)

    print("any-5")
    for seq in it.iter_sequences(5, method="any"):
        print(seq)

    print("any_prefix-5")
    for seq in it.iter_sequences(5, method="any_prefix"):
        print(seq)

    print("any_suffix-5")
    for seq in it.iter_sequences(5, method="any_suffix"):
        print(seq)

    print("Empty list, any")
    for seq in SimpleIterator([]).iter_sequences(5, method="any"):
        print(seq)

    it.reset()
    print("Step-2")
    for v in it.iter(step=2):
        print(v)
    it.reset()

    print("limit-2")
    for v in it.iter(limit=2):
        print(v)

    print("iter_sequences")
    for v in iter_sequences(list(range(5)), n=3, method="all"):
        print(v)

    print("iter_pairs")
    for v in iter_pairs(list(range(5)), method="any_prefix"):
        print(v)


if __name__ == "__main__":
    test_simple_bounded_iterator()



import time
from loguru import logger
from utils_ak.simple_microservice import run_listener_async
from utils_ak.job_orchestrator.worker.worker import run_worker


def test_microservice_worker(worker_cls, payload, run_listener=True):
    from utils_ak.loguru import configure_loguru_stdout

    configure_loguru_stdout("DEBUG")
    if run_listener:
        run_listener_async("monitor_in", message_broker=payload["message_broker"])
    time.sleep(2)
    run_worker(worker_cls, {"worker_id": "worker_id", "payload": payload})
    logger.info("Finished batch")


def test_microservice_worker_deployment(
    deployment_fn,
    controller,
    message_broker,
    run_listener=True,
):
    from utils_ak.loguru import configure_loguru_stdout

    configure_loguru_stdout("DEBUG")
    if run_listener:
        run_listener_async("monitor_in", message_broker=message_broker)
    import anyconfig

    deployment = anyconfig.load(deployment_fn)

    controller.stop(deployment["id"])
    controller.start(deployment)
    time.sleep(5)
    controller.stop(deployment["id"])



import logging


class WindowManager:
    def add(self, key, value):
        raise NotImplemented()


class BasicWindowManager:
    def __init__(self, window_generator):
        self.windows = {}  # {key: window}
        self.window_generator = window_generator

    def add(self, key, value):
        if key not in self.windows:
            self.windows[key] = self.window_generator(key)

        window = self.windows[key]
        if window.state == "closed":
            logging.warning(f"Trying to add value to a closed window {key}")

        window.add(value)

        # todo: try to close window if possible here?

    def close_if_possible(self):
        res = {
            k: w
            for k, w in self.windows.items()
            if w.state == "open" and w.is_closeable()
        }
        return {k: w.close() for k, w in res.items()}


if __name__ == "__main__":
    from utils_ak.streaming import ProcessingSessionWindow

    wm = BasicWindowManager(lambda key: ProcessingSessionWindow(gap=2))

    wm.add("key", "asdf")

    import time

    print(wm.close_if_possible())
    time.sleep(1)
    print(wm.close_if_possible())
    time.sleep(1)
    print(wm.close_if_possible())



import time
import asyncio
import inspect
import numpy as np

from dateutil import rrule
from datetime import datetime, timedelta
from utils_ak.time import cast_sec, cast_timedelta

TIME_EPS = 0.001


class CallbackTimer:
    """ Run callback function every interval. """

    def __init__(
        self,
        callback,
        interval=1.0,
        timer_type="right",
        n_times=None,
        counter_type="left",
        args=None,
        kwargs=None,
    ):
        """
        :param timer_type: str, "left" to start timing immediately, "right" to wait for callback() to finish
        """
        self.callback = callback
        self.interval = cast_sec(interval)
        self.last_call = 0
        self.timer_type = timer_type
        self.counter_type = counter_type
        self.args = args or tuple()
        self.kwargs = kwargs or {}
        self.n_times = n_times
        self.counter = 0

    @property
    def next_call(self):
        if self.n_times and self.counter >= self.n_times:
            # no further call
            return np.inf
        else:
            return self.last_call + self.interval

    def run_if_possible(self):
        if time.time() > self.next_call:
            if self.timer_type == "right":
                try:
                    self.run()
                except:
                    self.last_call = time.time()
                    raise
                else:
                    self.last_call = time.time()

            elif self.timer_type == "left":
                self.last_call = time.time()
                self.run()
            return True
        return False

    def run(self):
        if self.n_times and self.counter_type == "left":
            self.counter += 1
        res = self.callback(*self.args, **self.kwargs)
        if self.n_times and self.counter_type == "right":
            self.counter += 1
        return res

    async def run_if_possible_async(self):
        if time.time() > self.next_call:
            if self.timer_type == "right":
                try:
                    await self.run_async()
                except:
                    self.last_call = time.time()
                    raise
                else:
                    self.last_call = time.time()
            elif self.timer_type == "left":
                self.last_call = time.time()
                await self.run_async()
            return True
        return False

    async def run_async(self):
        if self.n_times and self.counter_type == "left":
            self.counter += 1

        if inspect.iscoroutinefunction(self.callback):
            res = await self.callback(*self.args, **self.kwargs)
        else:
            res = self.callback(*self.args, **self.kwargs)

        if self.n_times and self.counter_type == "right":
            self.counter += 1

        return res


class ScheduleTimer:
    """ Run callback function at schedule. """

    def __init__(
        self,
        callback,
        pattern=None,
        freq=None,
        run_on_init=False,
        args=None,
        kwargs=None,
    ):
        """
        :param pattern: (seconds) (minutes) (hours) (day) (month) (day-of-week) in cron-like manner: * 5 * * * * ...
        """
        self.callback = callback
        self.next_call = 0
        self.run_on_init = run_on_init

        self.args = args or tuple()
        self.kwargs = kwargs or {}

        if freq:
            # cast pattern by frequency
            if cast_timedelta(freq) >= timedelta(days=1):
                # freq is supported only for hour and less frequency
                raise Exception(
                    f"Too large frequency: {cast_sec(freq)}. Use frequency less than a day"
                )
            freq_sec = round(cast_sec(freq))

            if freq_sec <= 0:
                raise Exception("Bad frequency: {}".format(freq_sec))

            unit_secs = [1, 60, 3600, 3600 * 24]

            # find which unit is used
            unit_i = 0
            while True:
                if unit_secs[unit_i] <= freq_sec < unit_secs[unit_i + 1]:
                    break
                unit_i += 1

            if unit_secs[unit_i + 1] % freq_sec != 0:
                raise Exception("Bad frequency: {}".format(freq_sec))

            freq_units = int(freq_sec / unit_secs[unit_i])
            freq_vals = [
                i * freq_units for i in range(unit_secs[unit_i + 1] // freq_sec)
            ]
            freq_vals = map(str, freq_vals)

            pattern_vals = []
            for i in range(6):
                if i < unit_i:
                    pattern_vals.append("0")
                elif i == unit_i:
                    if freq_units != 1:
                        pattern_vals.append(",".join(freq_vals))
                    else:
                        pattern_vals.append("*")
                elif i > unit_i:
                    pattern_vals.append("*")
            self.pattern = " ".join(pattern_vals)
        else:
            self.pattern = pattern
        units = sec, min, hou, day, mon, dow = [
            self._parse_time_unit(unit) for unit in self.pattern.split()
        ]
        freqs = [
            rrule.SECONDLY,
            rrule.MINUTELY,
            rrule.HOURLY,
            rrule.DAILY,
            rrule.MONTHLY,
            rrule.DAILY,
        ]
        param_names = [
            "bysecond",
            "byminute",
            "byhour",
            "bymonthday",
            "bymonth",
            "byweekday",
        ]

        try:
            freq = freqs[units.index(None)]
        except:
            # not stars at all
            freq = rrule.YEARLY

        self.rrule_params = dict(zip(param_names, units))
        self.rrule_params["freq"] = freq

        self._update_next_call()

    def _parse_time_unit(self, unit):
        if unit == "*":
            return None
        unit = unit.split(",")
        return [int(x) for x in unit]

    def _update_next_call(self):
        # set next_call to closest
        self.last_call_dt = (
            datetime.fromtimestamp(self.next_call) if self.next_call else datetime.now()
        )
        next_call_lst = list(
            rrule.rrule(
                dtstart=self.last_call_dt + timedelta(seconds=1),
                until=self.last_call_dt + timedelta(seconds=1) + timedelta(days=366),
                count=1,
                **self.rrule_params,
            )
        )

        if not next_call_lst:
            raise Exception("Bad scheduler input!")

        self.next_call = next_call_lst[0].timestamp()

    def run_if_possible(self):
        if self.run_on_init or time.time() > self.next_call:
            self.run_on_init = False
            self.run()
            self._update_next_call()
            return True
        return False

    async def run_if_possible_async(self):
        if self.run_on_init or time.time() > self.next_call:
            self.run_on_init = False
            await self.run_async()
            self._update_next_call()

    def run(self):
        return self.callback(*self.args, **self.kwargs)

    async def run_async(self):
        if inspect.iscoroutinefunction(self.callback):
            return await self.callback(*self.args, **self.kwargs)
        else:
            return self.callback(*self.args, **self.kwargs)


class CallbackTimers(object):
    """ A builtin of callback timers. """

    def __init__(self):
        self.timers = []

    def add_timer(self, timer):
        self.timers.append(timer)

    def run_if_possible(self):
        if time.time() > self.next_call:
            for timer in self.timers:
                timer.run_if_possible()

    @property
    def next_call(self):
        return min([timer.next_call for timer in self.timers])

    def next_call_timeout(self):
        return max(self.next_call - time.time(), 0.0)

    def __len__(self):
        return len(self.timers)

    async def run_if_possible_async(self):
        if time.time() > self.next_call:
            for timer in self.timers:
                await timer.run_if_possible()


def test():
    def print_msg(msg=None):
        print("Callback:", datetime.now(), msg)

    print("Timer 1")
    timer1 = CallbackTimer(print_msg, 1)
    timer2 = CallbackTimer(print_msg, 1, args=("asdf",))
    for i in range(300):
        timer1.run_if_possible()
        timer2.run_if_possible()
        time.sleep(0.01)

    print("Timer 2")
    # run every second
    timer1 = ScheduleTimer(print_msg, "* * * * * *")
    timer2 = ScheduleTimer(print_msg, "* * * * * *", args=("asdf",))
    for i in range(300):
        timer1.run_if_possible()
        timer2.run_if_possible()
        time.sleep(0.01)

    print("Timer 3")
    # run every second
    timer = ScheduleTimer(print_msg, freq="2s", run_on_init=True)
    print(timer.pattern)
    N = 3 * 1000
    for i in range(N):
        timer.run_if_possible()
        time.sleep(1 / 10000)


def test_async():
    import numpy as np

    async def print_msg(msg=None):
        timeout = np.random.uniform(0, 1)
        await asyncio.sleep(timeout)
        print("Callback:", datetime.now(), msg, timeout)

        if timeout > 0.5:
            raise Exception("Test")

    timer = CallbackTimer(print_msg, 1.0)

    async def wrapper():
        while True:
            try:
                await timer.coroutine()
            except Exception as e:
                print(e)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.ensure_future(wrapper()))
    loop.close()


if __name__ == "__main__":
    test()
    test_async()



""" Pipeline is a callable, which runs consecutively an array of callables. """


class Pipeline(object):
    def __init__(self, array):
        self.array = array

        assert len(self.array) > 1

    def run(self, *args, **kwargs):
        res = self.array[0](*args, **kwargs)
        for pipe in self.array[1:]:
            res = pipe(res)
        return res

    def __call__(self, *args, **kwargs):
        self.run(*args, **kwargs)


if __name__ == "__main__":
    print(Pipeline([lambda y: y * 2, lambda x: x * 2, lambda x: x + 10]).run(15))



import decimal
import datetime
import os
import yaml
import json
import ujson

from io import StringIO

from utils_ak.coder.coder import Coder
from utils_ak.numeric import is_int, is_float
from utils_ak.builtin import delistify


# todo: move to ujson for encoding
# todo: support proper decoding


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_json"):
            return obj.to_json()
        elif isinstance(obj, decimal.Decimal):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return str(obj)
        elif isinstance(obj, datetime.date):
            return str(obj)
        elif isinstance(obj, datetime.time):
            return str(obj)
        elif is_int(obj):
            return int(obj)
        elif is_float(obj):
            return float(obj)
        else:
            return super().default(obj)


class JsonCoder(Coder):
    def __init__(self, encoding="utf-8", indent=None):
        self.encoding = encoding
        self.indent = indent

    def encode(self, obj):
        if not isinstance(obj, str):
            obj = json.dumps(obj, cls=JsonEncoder, ensure_ascii=False)

        if self.encoding:
            obj = obj.encode(self.encoding)

        return obj

    def decode(self, s):
        return ujson.loads(s)


json_coder = JsonCoder(encoding=None)


def cast_js(js_obj):
    return json_coder.encode(js_obj)


def cast_dict_or_list(obj, *args, **kwargs):
    if isinstance(obj, (dict, list)):
        return obj

    if isinstance(obj, bytes):
        try:
            obj = obj.decode("utf-8")
        except:
            raise Exception("Unknown bytes encoding")

    # load object from file if path exists
    if isinstance(obj, str):
        if os.path.exists(obj):
            with open(obj, "r") as f:
                obj = f.read()

        try:
            res = json_coder.decode(obj)
            assert isinstance(res, (dict, list))
            return res
        except:
            pass

        # todo: hardcode, put in different place
        try:
            res = delistify(list(yaml.load_all(StringIO(obj), Loader=yaml.FullLoader)))
            assert isinstance(res, (dict, list))
            return res
        except:
            pass

    raise Exception("Unknown type")


if __name__ == "__main__":
    from utils_ak.coder import test_coder

    test_coder(JsonCoder(), assert_equal=False)



import os
import anyconfig
import collections
import sys


def cast_config(obj, required=False):
    if obj is None:
        if not required:
            return {}
        else:
            raise Exception("Config required")
    elif isinstance(obj, collections.abc.Mapping):
        return obj
    elif isinstance(obj, str):
        if os.path.exists(obj):
            return anyconfig.load(obj)
        else:
            if required:
                raise Exception("Config file not found {}".format(obj))
            else:
                return {}
    elif isinstance(obj, list):
        res = {}
        for v in obj:
            anyconfig.merge(res, cast_config(v), ac_merge=anyconfig.MS_DICTS)
        return res
    else:
        raise Exception("Unknown config type")


def get_config(
    configs=None,
    require_local=False,
    global_configs=("common_config.yml", "secret_config.yml", "instance_config.yml"),
):
    local_config_fn = os.path.splitext(os.path.abspath(sys.argv[0]))[0] + ".yml"

    if not os.path.exists(local_config_fn):
        if require_local:
            raise Exception(f"Local config not found {local_config_fn}")
        local_config_fn = None

    res = []
    # add global configs
    for base_config in global_configs:
        for dirname in [os.getcwd(), os.path.dirname(os.path.abspath(sys.argv[0]))]:
            fn = os.path.join(dirname, base_config)
            if os.path.exists(fn):
                res.append(fn)
                break

    configs = configs or []
    res += configs

    # add local config
    res.append(local_config_fn)

    return cast_config(res)


def test():
    print(
        get_config(
            global_configs=(
                "sample_configs/common_config.yml",
                "sample_configs/secret_config.yml",
                "sample_configs/instance_config.yml",
            )
        )
    )


if __name__ == "__main__":
    test()



import random

from utils_ak.simple_microservice import SimpleMicroservice, run_listener_async
from utils_ak.loguru import configure_loguru_stdout


class Heartbeater(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__(f"Publisher {random.randint(0, 10 ** 6)}", *args, **kwargs)
        self.add_timer(
            self.publish,
            1.0,
            args=(
                "monitor",
                "my_topic",
            ),
            kwargs={"id": self.id},
        )


if __name__ == "__main__":
    configure_loguru_stdout()
    run_listener_async(
        "monitor",
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "monitor": {"endpoint": "tcp://localhost:5555", "type": "sub"}
                }
            },
        ),
    )
    Heartbeater(
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "monitor": {"endpoint": "tcp://localhost:5555", "type": "sub"}
                }
            },
        )
    ).run()



class ProxyBag(object):
    """ I'm a bag of proxies. I store them, I give them :) """

    def __init__(self, proxies):
        super().__init__()

        self.proxies = proxies

    def all(self):
        return self.proxies

    def without(self, forsaken_proxies):
        return filter(lambda proxy: proxy not in forsaken_proxies, self.proxies)

    def add(self, new_proxies: list):
        self.proxies += new_proxies



import anyconfig
import time
from utils_ak.loguru import logger, configure_loguru_stdout


def test_controller(controller_cls):
    configure_loguru_stdout("DEBUG")

    deployment = anyconfig.load("../../example/deployment.yml")
    ctrl = controller_cls()
    logger.info("Starting")
    ctrl.start(deployment)
    time.sleep(3)
    logger.info("Logs")
    ctrl.log(deployment["id"])
    logger.info("Stopping")
    ctrl.stop(deployment["id"])



import pandas as pd
import numpy as np

import os
import base64

from PIL import Image
from io import BytesIO

from utils_ak.os import *
from utils_ak.builtin import iter_get, remove_neighbor_duplicates

pd.set_option("display.max_colwidth", None)


def display_with_image_thumbnails(df, shape=None):
    from IPython.display import HTML, display

    def _get_thumbnail(path, shape=None):
        if shape is None:
            shape = (150, 150)
        i = Image.open(path)
        i.thumbnail(shape, Image.LANCZOS)
        return i

    def _image_base64(im, shape):
        if isinstance(im, str):
            im = _get_thumbnail(im, shape)
        with BytesIO() as buffer:
            im.save(buffer, "jpeg")
            return base64.b64encode(buffer.getvalue()).decode()

    def _image_formatter(im, shape):
        if type(im) != str:
            return str(im)
        possible_path = im
        if not os.path.exists(possible_path):
            return im

        ext = os.path.splitext(possible_path)[-1]

        if ext not in [".png", ".jpg"]:
            return im

        return '<img src="granular_storage:image/jpeg;base64,{}">'.format(
            _image_base64(im, shape)
        )

    columns = df.columns
    formatters = {col: lambda x: _image_formatter(x, shape) for col in columns}
    # displaying PIL.Image objects embedded in dataframe
    display(HTML(df.to_html(formatters=formatters, escape=False)))


def where(cond, x, y):
    """np.where analogue for dataframes with the same signature. Return elements, either from x or y, depending on condition.

    Parameters
    ----------
    cond: boolean condition
    cond_value: value to return when cond is True
    other: value to return when cond is False

    Returns
    -------
    pd.DataFrame
        np.where wrapped pandas dataframe
    """
    if isinstance(x, pd.DataFrame):
        df = x
    elif isinstance(y, pd.DataFrame):
        df = y
    elif isinstance(cond, pd.DataFrame):
        df = cond
    else:
        raise Exception("Either cond, x, y must be a dataframe")
    return pd.DataFrame(np.where(cond, x, y), index=df.index, columns=df.columns)


def series_equal(s1, s2, dtype="float", eps=1e-8):
    if dtype == "float":
        return s1.sub(s2).abs() < eps
    else:
        return s1 == s2


def find_row(df, row):
    """Find last occurence of row in df
    :df: pd.DataFrame
    :row: pd.DataFrame with unit length
    :return: int. Row number
    """
    if len(row) != 1:
        raise Exception("Row should be one line dataframe")

    if not np.all(df.columns == row.columns):
        raise Exception("Dataframe and row have different columns")

    # filter by index
    # time is considered discrete - no problem on rounding here
    df.index.name = "index"
    df = df.reset_index()
    tmp = df[df["index"] == row.index[0]].copy()
    tmp.pop("index")

    if len(tmp) == 0:
        return None

    # filter by values
    for col in tmp.columns:
        # todo: make better
        dtype = "float" if "float" in str(df.dtypes[0]) else None

        row_series = pd.Series(index=tmp[col].index)
        row_series[:] = row[col].iloc[0]

        tmp[col] = series_equal(tmp[col], row_series, dtype=dtype)
    """ 0    False
        1     True
        2    False
        3    False
        dtype: bool"""
    row_equal = tmp.all(axis=1)
    tmp = tmp[row_equal]

    if len(tmp) == 0:
        return None

    return tmp.index[-1]


def merge_by_columns(dfs):
    res_df = dfs[0].copy()
    for df in dfs[1:]:
        for key in df.columns:
            res_df[key] = df[key]
    return res_df


def merge(dfs, by=None, by_index=False, keep="last", sort_index=True):
    """
    :param dfs: list(`pd.DataFrame`)
    :param by: name of column or list of columns names. 'all' for all columns. 'columns' for merge_by_columns method
    :param by_index: include index or not (bool value)
    :param keep: 'last' or 'first'

    :return:
    """
    if not by and not by_index:
        raise TypeError("Either by or index should by non-trivial")

    if by is None:
        by = []
    elif by == "all":
        pass
    elif by == "columns":
        return merge_by_columns(dfs)
    elif isinstance(by, str):
        by = [by]
    elif isinstance(by, list):
        pass
    else:
        raise TypeError("By should be list, str or None")

    df = pd.concat(dfs, axis=0)

    masks = []
    if by == "all":
        masks.append(df)
    elif by:
        masks.append(df[by])

    if by_index:
        masks.append(pd.Series(df.index, index=df.index))
    mask = pd.concat(masks, axis=1)
    df = df[~mask.duplicated(keep=keep)]

    if sort_index:
        df = df.sort_index()
    return df


def find_gaps(index, gap):
    index = pd.Series(index, index=index)
    index_diff = index - index.shift(1)
    index_diff = index_diff[index_diff > gap]
    return index_diff


# todo: search for existing solutions
def pd_read(fn, index_column=None, **kwargs):
    ext = os.path.splitext(fn)[-1]
    if ".zip" in ext:
        ext = os.path.splitext(fn[:-4])[-1]
    df = getattr(pd, f"read_{ext[1:]}")(fn, **kwargs)
    if index_column:
        df = df.set_index(index_column)
    return df


def pd_write(df, fn, index_column=None, **kwargs):
    """NOTE: index is dropped when writing pandas dataframe."""
    if index_column:
        prev_index_column = df.index.name or "index"
        df = df.reset_index()
        df[index_column] = df.pop(prev_index_column)
    ext = os.path.splitext(fn)[-1]
    if ".zip" in ext:
        ext = os.path.splitext(fn[:-4])[-1]
        kwargs["compression"] = "zip"
    kwargs["index"] = False

    tmp_fn = fn + ".tmp"
    res = getattr(df, f"to_{ext[1:]}")(tmp_fn, **kwargs)
    if os.path.exists(fn):
        remove_path(fn)
    rename_path(tmp_fn, fn)
    return res


def mark_consecutive_groups(df, key, groups_key):
    cur_value = None
    cur_i = 0

    values = []
    for i, row in df.iterrows():
        if row[key] != cur_value:
            cur_i += 1
            cur_value = row[key]
        values.append(cur_i)
    df[groups_key] = values


def df_to_tree(df, recursive=True, delistify=False):
    df = df.copy()
    if len(df.columns) == 1:
        res = list(set(df[df.columns[0]].tolist()))
        if delistify and len(res) == 1:
            res = res[0]
        return res

    res = {}

    for value, grp in df.groupby(df.columns[0]):
        grp.pop(df.columns[0])
        if recursive:
            res[value] = df_to_tree(grp, recursive=True, delistify=delistify)
        else:
            res[value] = grp
    return res


def df_to_ordered_tree(df, recursive=True, prune_last=True):
    df = df.copy()
    if len(df.columns) == 1:
        res = df[df.columns[0]].tolist()
        if prune_last:
            res = remove_neighbor_duplicates(res)
        return res

    res = []
    pre_res = []

    col = df.columns[0]
    mark_consecutive_groups(df, col, "__key")

    for value, grp in df.groupby([col, "__key"]):
        grp.pop(col)
        grp.pop("__key")
        if recursive:
            child = df_to_ordered_tree(grp, recursive=True, prune_last=prune_last)
        else:
            child = grp
        pre_res.append((value[0], value[1], child))

    pre_res = list(
        sorted(
            pre_res,
            key=lambda obj: df[(df[col] == obj[0]) & (df["__key"] == obj[1])].iloc[0][
                "__key"
            ],
        )
    )
    return [(x[0], x[2]) for x in pre_res]


def test():
    df1 = pd.DataFrame.from_dict({"a": [1, 2, 3], "b": [4, 5, 6], "c": [1, 1, 1]})
    df1 = df1.set_index("c")
    df2 = pd.DataFrame.from_dict({"a": [3, 4, 5], "b": [7, 8, 9]})

    print(df1)
    print("-----")
    print(df2)

    print("----------")

    dfs = [df1, df2]

    print(merge(dfs, by_index=False, by="a"))
    print()
    print(merge(dfs, by_index=True, by="a"))
    print()
    print(merge(dfs, by_index=True, by=None))
    print()
    print(merge(dfs, by_index=True, by=None, keep="first"))
    print()
    print(merge(dfs, by_index=True, by="all", keep="first"))

    try:
        print(merge(dfs, by_index=False, by=None))
    except Exception as e:
        print(e)
    else:
        raise Exception("Should not happen")


def test_tree():
    df = pd.DataFrame(
        [
            ["A", "a", "a1"],
            ["A", "a", "a2"],
            ["A", "a", "a1"],
            ["B", "b", "b1"],
            ["B", "b", "b2"],
            ["A", "c", "c1"],
            ["A", "c", "c2"],
            ["A", "c", "c2"],
        ],
        columns=["col1", "col2", "col3"],
    )
    print(df)
    print(df_to_tree(df))
    print(df_to_ordered_tree(df))
    print(df_to_ordered_tree(df, prune_last=False))


if __name__ == "__main__":
    test()
    test_tree()



""" Handler is a callable objects with multiple callbacks, filters and formatters. Supports asyncio run"""

from utils_ak.builtin import listify, delistify
import inspect


# todo: optimize aio code! Execute all callbacks using wait or something!


def is_async(func):
    if inspect.iscoroutinefunction(func):
        return True
    elif hasattr(func, "is_async") and getattr(func, "is_async"):
        return True
    else:
        return False


class Handler(object):
    """Complex func wrapper. Computational graph with multiple filters, multiple formatters and multiple callbacks."""

    def __init__(self, callback=None, formatter=None, filter=None, reducer=None):
        """
        :param filter: callable or list of callable. Each filter should return True or False if rule is not specified (see Handler.add_filter for more)
        :param formatter: callable or list of callable. Each formatter should return args, kwargs tuple
        :param callback: callable or list of callable
        :return func return if func is single and list or returns of callbacks otherwise (a list)

        See __main__ for examples.
        """
        self.filters = []
        self.formatters = []
        self.callbacks = []

        self.add_filter(filter)
        self.add_formatter(formatter)
        self.add_callback(callback)
        self.reducer = reducer or delistify

    @property
    def is_async(self):
        return any(is_async(callback) for callback in self.callbacks)

    def add(self, callback=None, formatter=None, filter=None):
        self.add_formatter(formatter)
        self.add_filter(filter)
        self.add_callback(callback)

    def add_formatter(self, formatter=None):
        if formatter:
            self.formatters += listify(formatter)

    def add_filter(self, filter=None, rule=None):
        if filter:
            if rule:
                if isinstance(filter, list):
                    raise Exception("Filter should be callable if rule is specified. ")
                _filter = lambda *args, **kwargs: rule(filter(*args, **kwargs))
            else:
                _filter = filter
            self.filters += listify(_filter)

    def add_callback(self, callback=None):
        if callback:
            self.callbacks += listify(callback)

    def __call__(self, *args, **kwargs):
        for formatter in self.formatters:
            args, kwargs = formatter(*args, **kwargs)
        if all(_filter(*args, **kwargs) for _filter in self.filters):
            return self.reducer(
                [callback(*args, **kwargs) for callback in self.callbacks]
            )

    def call(self, *args, **kwargs):
        return self.__call__(*args, **kwargs)

    async def call_async(self, *args, **kwargs):
        for formatter in self.formatters:
            args, kwargs = formatter(*args, **kwargs)
        if all(_filter(*args, **kwargs) for _filter in self.filters):
            res = []
            for callback in self.callbacks:
                if is_async(callback):
                    if hasattr(callback, "aiocall"):
                        res.append(await callback.call_async(*args, **kwargs))
                    else:
                        res.append(await callback(*args, **kwargs))
                else:
                    res.append(callback(*args, **kwargs))
            return self.reducer(res)

    def has_coroutine(self):
        return any(inspect.iscoroutinefunction(callback) for callback in self.callbacks)

    def __repr__(self):
        vals = []
        vals += [f"Formatter: {formatter}" for formatter in self.formatters]
        vals += [f"Filter: {filter}" for filter in self.filters]
        vals += [f"Callback: {callback}" for callback in self.callbacks]
        return "\n".join(vals)

    def __str__(self):
        return self.__repr__()



import pandas as pd
import os
from utils_ak.pandas import pd_write, pd_read
from utils_ak.os import makedirs, list_files, remove_path
from utils_ak.time import cast_dt, cast_str, cast_datetime_series
from utils_ak.pandas import merge


class PandasSplitCombineETL:
    def __init__(self, path, key_func=None, prefix="", extension=".csv", merge_by=None):
        self.path = path
        self.extension = extension
        makedirs(path)
        self.key_func = key_func
        self.merge_by = merge_by
        self.prefix = prefix

    def _split(self, combined):
        assert self.key_func, "Key func not defined"
        df = combined
        df["_key"] = self.key_func(df)
        df.columns = [str(c) for c in df.columns]

        if not df.index.name:
            df.index.name = "index"

        df = df.reset_index()
        for key, split in df.groupby("_key"):
            yield key, split.drop(["_key"], axis=1)

    def _fn(self, key):
        values = [self.prefix, key]
        values = [v for v in values if v]
        return os.path.join(self.path, "-".join(values) + self.extension)

    def _load(self, key, split):
        fn = self._fn(key)
        if os.path.exists(fn):
            current_df = self._extract(key)
            if self.merge_by:
                split = merge([current_df, split], by=self.merge_by)
            else:
                split = pd.concat([current_df, split], axis=0)
        pd_write(split, fn, index=False)

    def _get_keys(self):
        fns = list_files(self.path, pattern="*" + self.extension, recursive=True)
        keys = [os.path.splitext(os.path.basename(fn))[0] for fn in fns]
        # remove prefix if needed
        keys = [
            key[len(self.prefix + "-") :] if key.startswith(self.prefix + "-") else key
            for key in keys
        ]
        return keys

    def _extract(self, key):
        df = pd_read(self._fn(key))
        index_column = df.columns[0]
        df[index_column] = cast_datetime_series(df[index_column])
        df.columns = [str(c) for c in df.columns]
        return df

    def _combine(self, splits_dic):
        """
        :param splits_dic: {key: split}
        """
        dfs = list(splits_dic.values())
        df = pd.concat(dfs, axis=0)
        index_column = df.columns[0]
        df = df.set_index(index_column)
        df = df.sort_index()
        return df

    def split_and_load(self, combined):
        for key, split in self._split(combined):
            self._load(key, split)

    def extract_and_combine(self):
        keys = self._get_keys()
        splits_dic = {key: self._extract(key) for key in keys}
        return self._combine(splits_dic)


def write_pandas_granular(
    df, path, key_func=None, prefix="", extension=".csv", merge_by=None
):
    return PandasSplitCombineETL(
        path, key_func, prefix, extension, merge_by
    ).split_and_load(df)


def read_pandas_granular(
    path, key_func=None, prefix="", extension=".csv", merge_by=None
):
    return PandasSplitCombineETL(
        path, key_func, prefix, extension, merge_by
    ).extract_and_combine()


def test_pandas_split_combine_etl():
    import time
    from utils_ak.time import cast_dt
    from utils_ak.os import remove_path

    df = pd.DataFrame(
        list(range(100)),
        index=pd.date_range(cast_dt("2020.01.01"), periods=100, freq="1d"),
    )

    etl = PandasSplitCombineETL(
        path="test-data/",
        extension=".parquet",
        key_func=lambda df: pd.Series(df.index, index=df.index).apply(
            lambda dt: cast_str(dt, "%Y%m")
        ),
    )

    etl.split_and_load(df)

    print(etl.extract_and_combine())

    remove_path("test-data/")


if __name__ == "__main__":
    test_pandas_split_combine_etl()



import pandas as pd
from utils_ak.dag import *

from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import *
from utils_ak.fluid_flow.actors.container import Container
from utils_ak.fluid_flow.calculations import *
from utils_ak.iteration import SimpleIterator

from functools import wraps


def switch(f):
    @wraps(f)
    def inner(self, *args, **kwargs):
        pipe_switch(self, self.current("in"), "in")
        pipe_switch(self, self.current("out"), "out")

        res = f(self, *args, **kwargs)

        pipe_switch(self, self.current("in"), "in")
        pipe_switch(self, self.current("out"), "out")
        return res

    return inner


class Queue(Actor, PipeMixin):
    def __init__(self, name, lines, break_funcs=None):
        super().__init__(name)
        self.lines = lines

        self.df = pd.DataFrame(
            index=["in", "out"], columns=["iterators", "break_funcs", "paused"]
        )
        self.df["iterator"] = [SimpleIterator(lines, 0) for _ in range(2)]
        self.df["break_func"] = self.default_break_func
        self.df["paused"] = False

        break_funcs = break_funcs or {}
        for orient, break_func in break_funcs.items():
            self.df.at[orient, "break_func"] = break_func

        self.breaks = []  # ts_beg, ts_end

    def default_break_func(self, old, new):
        return 0

    def current(self, orient):
        return self.df.at[orient, "iterator"].current()

    def inner_actors(self):
        return self.lines

    @switch
    def update_value(self, ts):
        for line in self.lines:
            line.update_value(ts)

        for orient in ["in", "out"]:
            if self.current(orient).is_limit_reached(orient):
                if self.df.at[orient, "paused"]:
                    continue
                old = self.current(orient)
                new = self.df.at[orient, "iterator"].next(
                    return_last_if_out=True, update_index=False
                )
                break_period = self.df.at[orient, "break_func"](old, new)

                if old != new:
                    if break_period:
                        self.breaks.append([ts, ts + break_period])
                        self.df.at[orient, "paused"] = True
                        self.add_event(
                            f"queue.resume.{self.id}",
                            ts + break_period,
                            {"orient": orient},
                        )
                    else:
                        pipe_switch(old, new, orient)
                        self.df.at[orient, "iterator"].next(
                            return_last_if_out=True, update_index=True
                        )

            # todo: del
            # print('Current', self.id, orient, self.current(orient), self.current(orient).is_limit_reached(orient), self.current(orient).containers['out'].df)

    @switch
    def on_resume(self, topic, ts, event):
        self.df.at[event["orient"], "paused"] = False
        old = self.current(event["orient"])
        new = self.df.at[event["orient"], "iterator"].next(
            return_last_if_out=True, update_index=True
        )
        pipe_switch(old, new, event["orient"])

    @switch
    def update_pressure(self, ts):
        for node in self.inner_actors():
            node.update_pressure(ts)

    @switch
    def update_speed(self, ts):
        for node in self.inner_actors():
            node.update_speed(ts)

    @switch
    def update_triggers(self, ts):
        for node in self.inner_actors():
            node.update_triggers(ts)

    def __str__(self):
        return f"Queue: {self.name}"

    def stats(self):
        return [[node.name, node.stats()] for node in self.lines]

    def display_stats(self):
        return [(node.name, node.display_stats()) for node in self.lines]

    def active_periods(self, orient="in"):
        return sum([node.active_periods(orient) for node in self.lines], [])

    def subscribe(self):
        super().subscribe()
        self.event_manager.subscribe(f"queue.resume.{self.id}", self.on_resume)

    def reset(self):
        super().reset()
        self.breaks = []



import re
import numpy as np
import pandas as pd
from dateutil.tz import tzutc
from dateutil.parser import parse as parse_date
from datetime import datetime, timedelta, timezone
from utils_ak.numeric import custom_round


# NOTE: slow
def parse_human_timestamp_re(hts, min_date_str="2000"):
    """
    :param hts: Human timestamp: 20180101/2018010112/201801011200/20180101120000/20180101120000123...
    :return:
    """
    ts_PATTERN = re.compile(r"(\d{4})(\d{2})(\d{2})(\d{2})?(\d{2})?(\d{2})?(\d+)?")
    hts = str(hts)

    if hts < min_date_str:
        raise Exception("Min date test failed")

    split = list(ts_PATTERN.match(hts).groups()[: int((len(hts) - 2) / 2)])

    # adjust microseconds
    if len(split) == 7:
        split[-1] = split[-1].ljust(6, "0")[:6]

    return datetime(*map(int, split))


def parse_human_timestamp(hts, min_date_str="2000"):
    """
    :param hts: Human timestamp: 20180101/2018010112/201801011200/20180101120000/20180101120000123...
    :return:
    """
    hts = str(hts)

    if hts < min_date_str:
        raise Exception("Min date test failed")

    slices = [
        slice(0, 4),
        slice(4, 6),
        slice(6, 8),
        slice(8, 10),
        slice(10, 12),
        slice(12, 14),
        slice(14, None),
    ][: int((len(hts) - 2) / 2)]

    split = [hts[sl] for sl in slices]

    # adjust microseconds
    if len(split) == 7:
        split[-1] = split[-1].ljust(6, "0")[:6]

    return datetime(*map(int, split))


def cast_hts(dt_obj):
    dt_obj = cast_datetime(dt_obj)
    return int(dt_obj.strftime("%Y%m%d%H%M%S%f")[:-3])


def cast_datetime(dt_obj, none_invariant=True):
    """
    :param dt_obj: datetime-like object: datetime.datetime or str
    :return: datetime

    is_utc =
    NOTE: This is slow
    """
    if isinstance(dt_obj, datetime):
        return dt_obj
    elif isinstance(dt_obj, str) and not is_freq(dt_obj):
        try:
            return parse_human_timestamp(dt_obj)
        except:
            pass

        # '01.08.2019' type
        search = re.search(r"(\d\d)\.(\d\d)\.(\d\d\d\d)", dt_obj)
        if search:
            d, m, y = search.groups()
            return datetime(int(y), int(m), int(d))

        # '01.08.20' type
        for pat in [
            r"^(\d\d)\.(\d\d)\.(\d\d)$",
            r"^\d(\d\d)\.(\d\d)\.(\d\d)$",
            r"^\d(\d\d)\.(\d\d)\.(\d\d)^\d",
            r"^(\d\d)\.(\d\d)\.(\d\d)\d",
        ]:
            search = re.search(pat, dt_obj)
            if search:
                d, m, y = search.groups()
                return datetime(int(y) + 2000, int(m), int(d))

        dt = parse_date(dt_obj)
        if dt and dt.tzinfo:
            dt = dt.astimezone(tzutc()).replace(tzinfo=None)
        return dt

    elif none_invariant and dt_obj is None:
        return None
    elif isinstance(dt_obj, (int, float, np.integer)):
        try:
            return parse_human_timestamp(dt_obj)
        except:
            pass

        try:
            return parse_date(str(dt_obj))
        except:
            pass

        try:
            return datetime.fromtimestamp(dt_obj, tz=timezone.utc).replace(tzinfo=None)
        except:
            pass

        return datetime.fromtimestamp(dt_obj / 1000, tz=timezone.utc).replace(
            tzinfo=None
        )
    else:
        raise Exception("Unknown datetime-like object type")


cast_dt = cast_datetime


def cast_timestamp(dt_obj):
    """
    :param dt_obj: naive datetime
    :return:
    """
    dt_obj = cast_datetime(dt_obj)
    # timestamp is always in utc!
    return dt_obj.replace(tzinfo=timezone.utc).timestamp()


cast_ts = cast_timestamp


def cast_mts(dt_obj):
    return cast_timestamp(dt_obj) * 1000


def cast_str(dt, format=None):
    if isinstance(dt, str):
        return dt
    elif isinstance(dt, datetime):
        if not format:
            return str(dt)
        return dt.strftime(format)
    else:
        raise Exception("Unsupported type")


def get_strptime_pattern(s):
    """
    :param s: str
    :return: get strptime pattern

    NOTE: be careful with microseconds. It is not handled properly
    """
    if len(s) > 20:
        raise Exception("Too big string")

    return "%Y%m%d%H%M%S%f"[: int(len(s) - 2)]


def cast_datetime_series(s):
    """
    :param s: a series of datetime-like objects with the same prototype.
    :return: a datetime series
    """

    sample = s.iloc[0]
    # process hts case
    try:
        parse_human_timestamp(sample)
    except:
        pass
    else:
        sample = str(sample)
        pattern = get_strptime_pattern(sample)
        s = s.astype(str)

        # 20 is for full %Y%m%D%H%M%S%f and 17 is for the same format, but when %f is replaced with 3 digits, not 6 as is by default
        if len(sample) > 20:
            # crop to microseconds
            s = s[:20]
        elif len(sample) >= 17:
            # add zeros for microseconds
            s = s + "0" * (20 - len(sample))
        return pd.to_datetime(s, format=pattern)

    if isinstance(sample, (int, np.integer)):
        # considered as timestamp: 1521193807
        int_part = str(sample).split(".")[0]
        # '1521193807'
        if len(int_part) == 10:
            return pd.to_datetime(s, unit="s")
        # '1521193807000'
        elif len(int_part) == 13:
            return pd.to_datetime(s, unit="ms")
        # '1521193807000000'
        elif len(int_part) == 16:
            return pd.to_datetime(s, unit="us")
        # '1521193807000000000'
        elif len(int_part) == 19:
            return pd.to_datetime(s, unit="ns")

    elif isinstance(sample, (float, np.float)):
        # considered as timestamp: 1521193807
        int_part = str(sample).split(".")[0]
        # '1521193807'
        if len(int_part) == 10:
            return s.apply(datetime.utcfromtimestamp)
        # '1521193807000'
        elif len(int_part) == 13:
            return (s / 1000).apply(datetime.utcfromtimestamp)
    return pd.to_datetime(s, infer_datetime_format=True)


def cast_datetime_many(lst):
    # NOTE: THIS CODE INFERS DATETIME FORMAT! In other words, all passed values should have the same format
    return cast_datetime_series(pd.Series(lst)).tolist()


cast_dt_series = cast_datetime_series


def cast_freq(td_obj, keys=None):
    """
    :param td: `datetime.timedelta`
    :return: string of {}D{}H{}T{}S format for `pandas.Grouper`
    """
    td = cast_timedelta(td_obj)
    keys = keys or ["d", "h", "m", "s"]
    d = td.days
    h = td.seconds // 3600
    m = (td.seconds // 60) % 60
    s = td.seconds - (3600 * h + 60 * m)
    vals = [d, h, m, s]

    res = ""
    for i in range(4):
        if vals[i] != 0:
            res += f"{vals[i]}{keys[i]}"
    return res


def parse_freq(freq_str, keys=None):
    keys = keys or ["d", "h", "m", "s"]
    ts_PATTERN = re.compile(r"^((\d+){})?((\d+){})?((\d+){})?((\d+){})?$".format(*keys))
    groups = ts_PATTERN.match(freq_str).groups()

    if len(groups) % 2 != 0:
        raise Exception("Bad input str")
    groups = groups[::2]

    period_secs = [3600 * 24, 3600, 60, 1]
    secs_dict = dict(zip(keys, period_secs))

    if not any(groups):
        raise Exception("Bad input str")

    secs = 0
    for group in groups:
        if not group:
            continue
        key = group[-1]
        val = int(group[:-1])
        if key not in secs_dict:
            raise Exception("Bad input str")
        secs += val * secs_dict[key]
    return timedelta(seconds=secs)


def is_freq(freq_str, keys=None):
    try:
        parse_freq(freq_str, keys=keys)
    except:
        return False
    return True


def cast_timedelta(td_obj):
    if isinstance(td_obj, timedelta):
        return td_obj
    elif isinstance(td_obj, (int, float, np.integer)):
        return timedelta(seconds=td_obj)
    elif isinstance(td_obj, str):
        try:
            return timedelta(seconds=int(td_obj))
        except:
            pass
        return parse_freq(td_obj)
    else:
        raise Exception("Unknown td_obj format")


cast_td = cast_timedelta


def cast_sec(td_obj):
    if isinstance(td_obj, (int, float)):
        return td_obj
    return cast_timedelta(td_obj).total_seconds()


def cast_dateoffset(td_obj):
    # https://pandas.pydata.org/pandas-docs/stable/timeseries.html
    return cast_freq(td_obj, keys=["D", "H", "T", "S"])


def round_datetime(dt_obj, td_obj, rounding="nearest_half_even"):
    ts = cast_timestamp(dt_obj)
    ts = custom_round(ts, cast_sec(td_obj), rounding)
    return cast_datetime(ts)


def test():
    print(cast_datetime(datetime.now()))
    print(cast_datetime("20180101"))
    print(cast_datetime(20180101))

    import time

    print(time.time(), cast_datetime(time.time()))
    print(cast_datetime(str(datetime.now())))

    dt1 = cast_datetime("2018-01-01 12:00:00")
    ts = cast_timestamp(dt1)
    dt2 = cast_datetime(ts)
    dt3 = cast_datetime(ts)
    print(dt1, dt2, dt3)
    print(parse_human_timestamp(20180101120000123123123))
    print(parse_human_timestamp(20180101120000123))

    print(datetime.now().strftime("%f"))
    print(cast_hts(datetime.now()))
    print(datetime.fromtimestamp(1500021000))
    print(datetime.utcfromtimestamp(1500021000))

    print(cast_datetime(20170720000000000))

    print(cast_datetime(1))
    print(parse_freq("4d1s"))
    print(cast_timedelta("4d5m"))
    print(cast_timedelta(300))
    print(cast_freq(300))
    print(cast_timedelta("5m"))

    print(cast_dt(1522783575000))

    print(cast_datetime("Thu May 31 19:34:16 2018 +0000"))
    print(cast_datetime("Thu May 31 19:34:16 2018 +0300"))

    print(round_datetime(cast_datetime("2018-01-01 12:00:05"), 300))
    print(round_datetime(cast_datetime("2018-01-01 12:00:05"), 300, "ceil"))
    print(round_datetime(cast_datetime("2018-01-01 12:00:05"), 300, "floor"))


if __name__ == "__main__":
    test()



import uuid
from utils_ak.dag import DAGNode


class Actor(DAGNode):
    def __init__(self, name=None):
        super().__init__()
        self.last_ts = None
        self.name = name
        self.id = str(uuid.uuid4())
        self.event_manager = None

    def set_event_manager(self, event_manager):
        self.event_manager = event_manager
        for actor in self.inner_actors():
            actor.set_event_manager(event_manager)

    def inner_actors(self):
        return []

    def add_event(self, topic, ts, event):
        self.event_manager.add_event(topic, ts, event)

    def update_last_ts(self, ts):
        self.last_ts = ts
        for actor in self.inner_actors():
            actor.update_last_ts(ts)

    def subscribe(self):
        for node in self.inner_actors():
            node.subscribe()

    def stats(self):
        return {}

    def display_stats(self):
        return {}

    def active_periods(self):
        return []

    def reset(self):
        self.last_ts = None
        for actor in self.inner_actors():
            actor.reset()



""" Broker is a generalization of message queue communication. """
from .broker import Broker
from .zmq_broker import ZMQBroker
from .kafka_broker import KafkaBroker

from .cast import cast_message_broker



from dynaconf import Dynaconf, Validator

settings = Dynaconf(
    settings_files=[
        "configs/default_settings.yaml",
        "configs/settings.yaml",
        "configs/settings.local.yaml",
        "configs/.secrets.yaml",
    ],
    environments=True,
    load_dotenv=True,
    envvar_prefix="DYNACONF",
    env_switcher="ENV_FOR_DYNACONF",
    dotenv_path="configs/.env",
)



class IntegrityGuard(object):
    """
    I'm an abstraction. Extend me.

    I check if some operation has gone good or has crashed anywhere in the past.
    Usually, I raise some flags before entering the unstable area, and put it down upon leaving it.
    If I find a risen flag -- we've crashed at some point and need to take some actions.
    Those actions are out of our scope -- we just check.
    """

    def is_integral(self):  # Are we integral at the moment? Is there a risen flag?
        raise NotImplementedError()

    def enter_unstable_state(
        self,
    ):  # Boom, we are unstable now. integral() must be false after calling this
        raise NotImplementedError()

    def leave_unstable_state(self):  # Hurray, we're done, kill the flag!
        raise NotImplementedError()



import numpy as np

ERROR = 1e-5


def nanmin(arr, require_any=True):
    arr = [v if v is not None else np.nan for v in arr]
    if require_any and all(np.isnan(v) for v in arr):
        raise Exception("No values found")
    return float(np.nanmin(arr))



import fire
from functools import partial
from utils_ak.job_orchestrator.worker.test.test_worker import TestWorker
from utils_ak.job_orchestrator.worker.worker import *
from utils_ak.loguru import configure_loguru_stdout

if __name__ == "__main__":
    configure_loguru_stdout()
    fire.Fire(partial(run_worker, worker_cls=TestWorker))



import os
import time
import fire
from loguru import logger
from notifiers import get_notifier
from datetime import datetime

# todo: make beep properly

from utils_ak.deployment.example.config import settings


def main(name=None, run_forever=True, beep=True):
    telegram = get_notifier("telegram")

    name = name or os.environ.get("NAME") or "World"
    logger.info(f"Hello {name}!")

    if beep and settings.get("telegram_bot_token"):
        for i in range(5):
            telegram.notify(
                message=f"Hi! from {datetime.now()}",
                token=settings["telegram_bot_token"],
                chat_id=settings["telegram_chat_id"],
            )
            time.sleep(2)

    if run_forever:
        while True:
            time.sleep(2)


if __name__ == "__main__":
    fire.Fire(main)



# import time
# import zmq
# from datetime import datetime
#
# from multiprocessing import Process
#
# from utils_ak.message_queue.brokers import BrokerManager
#
# SYSTEM_URL = 'tcp://127.0.0.1:5557'
# POLL = True
# # copy=True is much faster
# COPY = True
# SIZE = 1024
# COUNT = 100_000
# # lat or thr
# now = time.perf_counter
#
# # 'kafka' or 'zmq'
# BROKER = 'kafka'
#
# brokers_config = {
#     'zmq': {
#         'endpoints': {'benchmark': {'type': 'sub', 'endpoint': 'tcp://127.0.0.1:5556'}}
#     }
# }
#
#
# def throughput_sink():
#     context = zmq.Context()
#     # router_system - server analogue
#     router_system = context.socket(zmq.ROUTER)
#
#     broker = BrokerManager(BROKER, brokers_config)
#
#     broker.subscribe('benchmark', '')
#
#     router_system.bind(SYSTEM_URL)
#
#     time.sleep(0.2)
#
#     system_msg = router_system.recv_multipart()
#     assert system_msg[1] == b'BEGIN', system_msg
#
#     count = int(system_msg[2].decode('ascii'))
#     router_system.send_multipart(system_msg)
#
#     counter = 0
#     while counter != count:
#         msg = broker.poll(0.01)
#         if msg:
#             counter += 1
#
#     # system_msg[0] - instance to which we send message!
#     router_system.send_multipart([system_msg[0], b'DONE'])
#
#     router_system.close()
#     context.term()
#
#
# def get_throughput():
#     context = zmq.Context()
#     # dealer - client analogue
#     system_dealer = context.socket(zmq.DEALER)
#
#     broker = BrokerManager(BROKER, brokers_config)
#
#     # no memory limits for sending
#     system_dealer.SNDHWM = 0
#
#     #  Add your socket options here.
#     #  For example ZMQ_RATE, ZMQ_RECOVERY_IVL and ZMQ_MCAST_LOOP for PGM.
#
#     system_dealer.connect(SYSTEM_URL)
#
#     time.sleep(0.2)
#     data = b' ' * SIZE
#     # granular_storage = str(datetime.now())
#
#     system_dealer.send_multipart([b'BEGIN', str(COUNT).encode('ascii')])
#     # Wait for the other side to connect.
#     msg = system_dealer.recv_multipart()
#     assert msg[0] == b'BEGIN'
#     time.sleep(0.1)
#     start = now()
#     for i in range(COUNT):
#         broker.publish('benchmark', '', data)
#     sent = now()
#     # wait for receiver
#     reply = system_dealer.recv_multipart()
#     elapsed = now() - start
#     assert reply[0] == b'DONE'
#     send_only = sent - start
#
#     send_throughput = COUNT / send_only
#     throughput = COUNT / elapsed
#     megabits = throughput * SIZE * 8 / 1e6
#
#     print('Throughput')
#     print(f'Message size   : {SIZE}     [B]')
#     print(f'Message count  : {COUNT}     [msgs]')
#     print(f'Send only      : {send_throughput}     [msg/s]')
#     print(f'Mean throughput: {throughput}     [msg/s]')
#     print(f'Mean throughput: {megabits} [Mb/s]')
#     print(f'Test time      : {elapsed} [s]')
#     print()
#     context.destroy()
#     return (send_throughput, throughput)
#
#
# def benchmark_throughput():
#     p = Process(target=throughput_sink)
#     p.start()
#     get_throughput()
#     p.join()
#
#
# if __name__ == '__main__':
#     benchmark_throughput()



from utils_ak.fluid_flow import *
from utils_ak.block_tree import *

import warnings

warnings.filterwarnings("ignore")


def test_water_line_flow_1():
    drenator = Container("Drenator", value=1000, max_pressures=[None, None])

    melting1 = Processor(
        "Melting1",
        items=["a", "a"],
        max_pressures=[1000, 1000],
        processing_time=0,
        limits=[750, 750],
    )
    melting2 = Processor(
        "Melting2",
        items=["b", "b"],
        max_pressures=[1000, 1000],
        processing_time=0,
        limits=[250, 250],
    )
    melting_queue = Queue(
        "MeltingQueue", [melting1, melting2], break_funcs={"in": lambda old, new: 1}
    )

    cooling1 = Processor(
        "Cooling1",
        items=["a", "a"],
        max_pressures=[None, None],
        processing_time=0.5,
        limits=[750, 750],
    )
    cooling2 = Processor(
        "Cooling2",
        items=["b", "b"],
        max_pressures=[None, None],
        processing_time=0.5,
        limits=[250, 250],
    )
    cooling_queue = Queue("CoolingQueue", [cooling1, cooling2])

    packing_hub = Hub("Hub")

    packing1 = Processor(
        "Packing1",
        items=["a", "a1"],
        max_pressures=[200, None],
        processing_time=0,
        limits=[750, None],
    )
    packing2 = Processor(
        "Packing2",
        items=["b", "b1"],
        max_pressures=[400, None],
        processing_time=0,
        limits=[100, None],
    )
    packing_queue1 = Queue("PackingQueue1", [packing1, packing2])

    # using sequence as packing
    container = Container(
        "Container3", item="b", max_pressures=[200, None], limits=[150, None]
    )
    # slow packing...
    processor = Processor(
        "Processor3", items=["b", "b2"], max_pressures=[50, None], processing_time=0
    )
    packing3 = Sequence("Packing3", [container, processor])
    packing_queue2 = Queue("PackingQueue1", [packing3])

    pipe_connect(drenator, melting_queue, "drenator-melting")
    pipe_connect(melting_queue, cooling_queue, "melting-cooling")
    pipe_connect(cooling_queue, packing_hub, "cooling-hub")

    pipe_connect(packing_hub, packing_queue1, "hub-packing_queue1")
    pipe_connect(packing_hub, packing_queue2, "hub-packing_queue2")

    flow = FluidFlow(drenator, verbose=True)
    run_flow(flow)

    maker, make = init_block_maker("root", axis=1)

    for node in drenator.iterate("down"):
        if node.active_periods():
            for period in node.active_periods():
                label = "-".join([str(node.name), period[0]])
                beg, end = period[1:]
                beg, end = (
                    custom_round(beg * 60, 5) // 5,
                    custom_round(end * 60, 5) // 5,
                )
                make(label, x=[beg, 0], size=(end - beg, 1))
    print(maker.root.tabular())


def test_water_line_flow_2_multiple_boilings():
    n_boilings = 2
    packing1_1 = Processor(
        "Packing1",
        items=["a", "a1"],
        max_pressures=[200, None],
        processing_time=0,
        limits=[750, None],
    )
    packing2_1 = Processor(
        "Packing2",
        items=["b", "b1"],
        max_pressures=[400, None],
        processing_time=0,
        limits=[100, None],
    )
    packing1_2 = Processor(
        "Packing1",
        items=["a", "a1"],
        max_pressures=[200, None],
        processing_time=0,
        limits=[750, None],
    )
    packing2_2 = Processor(
        "Packing2",
        items=["b", "b1"],
        max_pressures=[400, None],
        processing_time=0,
        limits=[100, None],
    )
    packing_queue1 = Queue(
        "PackingQueue1", [packing1_1, packing2_1, packing1_2, packing2_2]
    )

    packing3_1 = Processor(
        "Packing3",
        items=["b", "b2"],
        max_pressures=[200, None],
        processing_time=0,
        limits=[150, None],
    )
    packing3_2 = Processor(
        "Packing3",
        items=["b", "b2"],
        max_pressures=[200, None],
        processing_time=0,
        limits=[150, None],
    )
    packing_queue2 = Queue("PackingQueue1", [packing3_1, packing3_2])

    for _ in range(n_boilings):
        drenator = Container("Drenator", value=1000, max_pressures=[None, None])

        melting1 = Processor(
            "Melting1",
            items=["a", "a"],
            max_pressures=[1000, 1000],
            processing_time=0,
            limits=[750, 750],
        )
        melting2 = Processor(
            "Melting2",
            items=["b", "b"],
            max_pressures=[1000, 1000],
            processing_time=0,
            limits=[250, 250],
        )
        melting_queue = Queue(
            "MeltingQueue", [melting1, melting2], break_funcs={"in": lambda old, new: 1}
        )

        cooling1 = Processor(
            "Cooling1",
            items=["a", "a"],
            max_pressures=[None, None],
            processing_time=0.5,
            limits=[750, 750],
        )
        cooling2 = Processor(
            "Cooling2",
            items=["b", "b"],
            max_pressures=[None, None],
            processing_time=0.5,
            limits=[250, 250],
        )
        cooling_queue = Queue("CoolingQueue", [cooling1, cooling2])

        packing_hub = Hub("Hub")

        pipe_connect(drenator, melting_queue, "drenator-melting")
        pipe_connect(melting_queue, cooling_queue, "melting-cooling")
        pipe_connect(cooling_queue, packing_hub, "cooling-hub")

        pipe_connect(packing_hub, packing_queue1, "hub-packing_queue1")
        pipe_connect(packing_hub, packing_queue2, "hub-packing_queue2")

        flow = FluidFlow(drenator, verbose=True)
        run_flow(flow)

        maker, make = init_block_maker("root", axis=1)

        for node in drenator.iterate("down"):
            if node.active_periods():
                for period in node.active_periods():
                    label = "-".join([str(node.name), period[0]])
                    beg, end = period[1:]
                    beg, end = (
                        custom_round(beg * 60, 5) // 5,
                        custom_round(end * 60, 5) // 5,
                    )
                    make(label, x=[beg, 0], size=(end - beg, 1))
        print(maker.root.tabular())

        # clean-up
        for node in drenator.iterate("down"):
            node.reset()

        # remove packers from current boiling
        pipe_disconnect(packing_hub, packing_queue1)
        pipe_disconnect(packing_hub, packing_queue2)


if __name__ == "__main__":
    test_water_line_flow_1()
    # test_water_line_flow_2_multiple_boilings()



import time
import zmq

from multiprocessing import Process

URL = "tcp://127.0.0.1:5555"
POLL = False
# copy=True is much faster
COPY = True
SIZE = 1024
COUNT = 1_000_000
# lat or thr
now = time.perf_counter

# If copy is unspecified (False in our case?), messages smaller than this many bytes will be copied and messages larger than this will be shared with libzmq.
# Should the frame(s) be sent in a copying or non-copying manner. If copy=False, frames smaller than self.copy_threshold bytes will be copied anyway.
# matters only for copy=false in my understanding!
zmq.COPY_THRESHOLD = 0


# docs
# socket types: http://api.zeromq.org/4-1:zmq-socket
# some options dock. http://api.zeromq.org/2-1:zmq-setsockopt


def latency_echo():
    """echo messages on a REP socket. Should be started before `latency` """
    context = zmq.Context()
    socket = context.socket(zmq.REP)

    if POLL:
        poller = zmq.Poller()
        poller.register(socket)

    socket.bind(URL)

    block = zmq.NOBLOCK if POLL else 0

    for i in range(COUNT + 1):
        if POLL:
            res = poller.poll()
        msg = socket.recv(block, copy=COPY)

        if POLL:
            res = poller.poll()

        socket.send(msg, block, copy=COPY)

    msg = socket.recv()
    assert msg == b"done"

    socket.close()
    context.term()


def get_latency():
    """Perform a latency test"""
    context = zmq.Context()
    s = context.socket(zmq.REQ)
    # behaviour on socket close and term
    s.setsockopt(zmq.LINGER, -1)
    s.connect(URL)

    if POLL:
        poller = zmq.Poller()
        poller.register(s)

    msg = b" " * SIZE

    block = zmq.NOBLOCK if POLL else 0
    # trigger one roundtrip before starting the timer
    s.send(msg)
    s.recv()
    start = now()

    for i in range(COUNT):
        if POLL:
            res = poller.poll()
            assert res[0][1] & zmq.POLLOUT
        s.send(msg, block, copy=COPY)

        if POLL:
            res = poller.poll()
            assert res[0][1] & zmq.POLLIN
        msg = s.recv(block, copy=COPY)

        assert len(msg) == SIZE

    elapsed = now() - start

    s.send(b"done")

    latency = 1e6 * elapsed / (COUNT * 2.0)

    print("Latency")
    print(f"Message size   : {SIZE}     [B]")
    print(f"Roundtrip count: {COUNT}     [msgs]")
    print(f"Mean latency   : {latency} [µs]")
    print(f"Test time      : {elapsed} [s]")
    print()

    context.destroy()
    return latency


def throughput_sink():
    context = zmq.Context()
    # router - server analogue
    socket = context.socket(zmq.ROUTER)
    # no memory limits for receiving
    socket.RCVHWM = 0

    #  Add your socket options here.
    #  For example ZMQ_RATE, ZMQ_RECOVERY_IVL and ZMQ_MCAST_LOOP for PGM.

    if POLL:
        poller = zmq.Poller()
        poller.register(socket)

    socket.bind(URL)
    msg = socket.recv_multipart()
    assert msg[1] == b"BEGIN", msg
    count = int(msg[2].decode("ascii"))
    socket.send_multipart(msg)

    flags = zmq.NOBLOCK if POLL else 0

    for i in range(count):
        if POLL:
            res = poller.poll()
            assert res[0][1] & zmq.POLLIN
        msg = socket.recv_multipart(flags=flags, copy=COPY)

    # msg[0] - instance to which we send message!
    socket.send_multipart([msg[0], b"DONE"])

    socket.close()
    context.term()


def get_throughput():
    context = zmq.Context()
    # dealer - client analogue
    socket = context.socket(zmq.DEALER)
    # no memory limits for sending
    socket.SNDHWM = 0

    #  Add your socket options here.
    #  For example ZMQ_RATE, ZMQ_RECOVERY_IVL and ZMQ_MCAST_LOOP for PGM.

    if POLL:
        poller = zmq.Poller()
        poller.register(socket, zmq.POLLOUT)

    socket.connect(URL)
    data = b" " * SIZE

    flags = zmq.NOBLOCK if POLL else 0
    socket.send_multipart([b"BEGIN", str(COUNT).encode("ascii")])
    # Wait for the other side to connect.
    msg = socket.recv_multipart()
    assert msg[0] == b"BEGIN"
    start = now()
    for i in range(COUNT):
        if POLL:
            res = poller.poll()
            assert res[0][1] & zmq.POLLOUT
        socket.send(data, flags=flags, copy=COPY)
    sent = now()
    # wait for receiver
    reply = socket.recv_multipart()
    elapsed = now() - start
    assert reply[0] == b"DONE"
    send_only = sent - start

    send_throughput = COUNT / send_only
    throughput = COUNT / elapsed
    megabits = throughput * SIZE * 8 / 1e6

    print("Throughput")
    print(f"Message size   : {SIZE}     [B]")
    print(f"Message count  : {COUNT}     [msgs]")
    print(f"Send only      : {send_throughput}     [msg/s]")
    print(f"Mean throughput: {throughput}     [msg/s]")
    print(f"Mean throughput: {megabits} [Mb/s]")
    print(f"Test time      : {elapsed} [s]")
    print()
    context.destroy()
    return (send_throughput, throughput)


def benchmark_throughput():
    p = Process(target=throughput_sink)
    p.start()
    get_throughput()
    p.join()


def benchmark_latency():
    p = Process(target=latency_echo)
    p.start()
    get_latency()
    p.join()


if __name__ == "__main__":
    benchmark_throughput()
    benchmark_latency()



""" Operation is an atomic function. Operation classes provide functionality to process failures and to be ensured that operation has finished or not successfully. """
from .operation import AsyncioTaskOperation



""" Python exceptions traceback functionality. """
import sys
import traceback


# https://stackoverflow.com/questions/13210436/get-full-traceback


class FauxTb(object):
    def __init__(self, tb_frame, tb_lineno, tb_next):
        self.tb_frame = tb_frame
        self.tb_lineno = tb_lineno
        self.tb_next = tb_next


def get_current_stack(skip=0):
    try:
        1 / 0
    except ZeroDivisionError:
        f = sys.exc_info()[2].tb_frame
    for i in range(skip + 2):
        f = f.f_back
    lst = []
    while f is not None:
        lst.append((f, f.f_lineno))
        f = f.f_back
    return lst


def extend_traceback(tb, stack):
    """Extend traceback with stack info."""
    head = tb
    for tb_frame, tb_lineno in stack:
        head = FauxTb(tb_frame, tb_lineno, head)
    return head


def full_exc_info():
    """Like sys.exc_info, but includes the full traceback."""
    t, v, tb = sys.exc_info()
    full_tb = extend_traceback(tb, get_current_stack(1))
    return t, v, full_tb


def test():
    def func():
        try:
            raise Exception("Dummy")
        except Exception as e:
            t, v, full_tb = full_exc_info()
            print(t)
            print(v)
            print(full_tb)
            print(traceback.format_exc())

    def func2():
        func()

    func2()


if __name__ == "__main__":
    test()



from mongoengine import *

from datetime import datetime


class Job(Document):
    AUTO_FIELDS = ["created", "_id"]
    type = StringField(required=True)
    payload = DictField()

    created = DateTimeField(default=datetime.utcnow)

    locked_by = ReferenceField("Worker")
    locked_at = DateTimeField()

    runnable = DictField(required=True)
    params = DictField()

    running_timeout = IntField()
    initializing_timeout = IntField(default=20.0, required=True)

    status = StringField(
        required=True,
        default="pending",
        choices=[
            "pending",
            "initializing",
            "running",
            "error",
            "success",
            "stalled",
            "terminated",
        ],
    )

    workers = ListField(ReferenceField("Worker"))

    meta = {"allow_inheritance": True}


class Worker(Document):
    AUTO_FIELDS = ["created", "_id"]
    job = ReferenceField(Job)
    config = DictField()
    created = DateTimeField(default=datetime.utcnow)
    status = StringField(
        required=True,
        default="pending",
        choices=[
            "pending",
            "initializing",
            "running",
            "error",
            "success",
            "stalled",
            "terminated",
        ],
    )
    response = StringField()

    meta = {"allow_inheritance": True}



import psutil

import subprocess
from utils_ak.str import cast_unicode

import re

import pandas as pd


def get_system_usage_stats():
    # todo: add cpu stats
    stats = {}

    # svmem(total=17058336768, available=6641549312, percent=61.1, used=10416787456, free=6641549312)
    memory = psutil.virtual_memory()
    stats["ram"] = {
        "total": memory.total,
        "available": memory.available,
        "free": memory.free,
        "percent": memory.percent,
    }

    # sdiskusage(total=254277836800, used=212004134912, free=42273701888, percent=83.4)
    disk = psutil.disk_usage("/")
    stats["disk"] = {
        "total": disk.total,
        "used": disk.used,
        "free": disk.free,
        "percent": disk.percent,
    }
    return stats


def get_dfh():
    output = execute("df -h | grep -v 'map '")  # grep -v 'map ' for mac book
    s = output[0]
    s = s.replace("Mounted on", "Mounted")
    s = s.replace("%iused", "Use%")  # for mac
    lines = s.strip().split("\n")
    lines = [re.sub("\s+", " ", line) for line in lines]
    values = [line.split(" ") for line in lines]
    df = pd.DataFrame(values[1:], columns=values[0])
    return df


# todo: tests



"""Declare basic string types unambiguously for various Python versions.
Authors
-------
* MinRK
"""

# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import sys

if sys.version_info[0] >= 3:
    bytes = bytes
    unicode = str
    basestring = (bytes, unicode)
else:
    unicode = unicode
    bytes = str
    basestring = basestring


def cast_bytes(s, encoding="utf8", errors="strict"):
    """cast unicode or bytes to bytes"""
    if isinstance(s, bytes):
        return s
    elif isinstance(s, unicode):
        return s.encode(encoding, errors)
    else:
        raise TypeError("Expected unicode or bytes, got %r" % s)


def cast_unicode(s, encoding="utf8", errors="strict"):
    """cast bytes or unicode to unicode"""
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    elif isinstance(s, unicode):
        return s
    else:
        raise TypeError("Expected unicode or bytes, got %r" % s)


# give short 'b' alias for cast_bytes, so that we can use fake b('stuff')
# to simulate b'stuff'
b = asbytes = cast_bytes
u = cast_unicode

__all__ = [
    "asbytes",
    "bytes",
    "unicode",
    "basestring",
    "b",
    "u",
    "cast_bytes",
    "cast_unicode",
]



from collections import OrderedDict


class LastUsedOrderedDict(OrderedDict):
    """Store items in the order the keys were last added"""

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        super().__setitem__(key, value)

    def pick(self, key):
        self.__setitem__(key, self[key])
        return self[key]


def test():
    d = LastUsedOrderedDict()
    d["1"] = 1
    d["2"] = 2
    print(d)
    print(d.pick("1"))
    print(d)
    d["2"] = 3
    print(d)


if __name__ == "__main__":
    test()



import os
import openpyxl as opx
from openpyxl.styles import Alignment, PatternFill, Font
from openpyxl.styles.borders import Border, Side, BORDER_THIN
from openpyxl.utils import get_column_letter
from utils_ak.color import cast_color


def init_workbook(sheet_names=None, active_sheet_name=None):
    workbook = opx.Workbook()
    sheet_names = sheet_names or []
    for i, sheet_name in enumerate(sheet_names):
        workbook.create_sheet(sheet_name, i)
    workbook.remove(workbook.worksheets[-1])
    if active_sheet_name:
        workbook.active = sheet_names.index(active_sheet_name)
    return workbook


def cast_workbook(wb_obj):
    if isinstance(wb_obj, str):
        return opx.load_workbook(filename=wb_obj, data_only=True)
    elif isinstance(wb_obj, opx.Workbook):
        return wb_obj
    elif isinstance(wb_obj, list):
        return init_workbook(sheet_names=wb_obj)
    else:
        raise Exception("Unknown workbook format")


def set_border(sheet, x, y, w, h, border):
    rows = sheet[
        "{}{}".format(get_column_letter(x), y) : "{}{}".format(
            get_column_letter(x + w - 1), y + h - 1
        )
    ]

    for row in rows:
        row[0].border = Border(
            left=border,
            top=row[0].border.top,
            bottom=row[0].border.bottom,
            right=row[0].border.right,
        )
        row[-1].border = Border(
            left=row[-1].border.left,
            top=row[-1].border.top,
            bottom=row[-1].border.bottom,
            right=border,
        )
    for c in rows[0]:
        c.border = Border(
            left=c.border.left, top=border, bottom=c.border.bottom, right=c.border.right
        )
    for c in rows[-1]:
        c.border = Border(
            left=c.border.left, top=c.border.top, bottom=border, right=c.border.right
        )


def draw_cell(
    sheet, x, y, text, color=None, font_size=None, text_rotation=None, alignment=None
):
    cell = sheet.cell(row=y, column=x)
    cell.font = Font(size=font_size)
    if alignment == "center":
        cell.alignment = Alignment(
            horizontal="center",
            vertical="center",
            wrap_text=True,
            text_rotation=text_rotation,
        )
    cell.value = text

    if color:
        cell.fill = PatternFill("solid", fgColor=color[1:])
    return cell


def draw_block(
    sheet,
    x1,
    x2,
    h1,
    h2,
    text,
    color=None,
    border=None,
    text_rotation=None,
    font_size=None,
    alignment=None,
    bold=False,
):
    color = color or cast_color("white")
    sheet.merge_cells(
        start_row=x2, start_column=x1, end_row=x2 + h2 - 1, end_column=x1 + h1 - 1
    )
    merged_cell = sheet.cell(row=x2, column=x1)
    merged_cell.font = Font(size=font_size, bold=bold)
    if alignment == "center":
        merged_cell.alignment = Alignment(
            horizontal="center",
            vertical="center",
            wrap_text=True,
            text_rotation=text_rotation,
        )
    merged_cell.value = text
    if color:
        merged_cell.fill = PatternFill("solid", fgColor=color[1:])

    if border is not None:
        if isinstance(border, dict):
            border = Side(**border)
        elif isinstance(border, Side):
            pass
        else:
            raise Exception("Unknown border type")

        set_border(sheet, x1, x2, h1, h2, border)


def draw_row(sheet, y, values, color=None, **kwargs):
    for i, v in enumerate(values, 1):
        draw_cell(sheet, i, y, text=v, color=color, **kwargs)


if __name__ == "__main__":
    wb = init_workbook(["a", "b"], active_sheet_name="b")
    wb.save("text.xlsx")



from utils_ak.architecture import TopicHandler, PrefixHandler, Handler


def test():
    def callback1(msg):
        print("callback1", msg)
        return 1

    def callback2(msg):
        print("callback2", msg)
        return 2

    def filter(msg):
        print("filter", msg)
        return "bar1" in msg

    def formatter(msg):
        return (msg.split(".")[0],), {}

    handler = Handler(callback=callback1)
    print("Test 1")
    print(handler("foo.bar1"))
    print(handler("foo.bar2"))
    print()

    handler = Handler(callback=[callback1, callback2], filter=filter)
    print("Test 2")
    print(handler("foo.bar1"))
    print(handler("foo.bar2"))
    print()

    handler = Handler(callback=[callback1, callback2])
    handler.add_filter(filter, rule=lambda b: not b)
    print("Test 3")
    print(handler("foo.bar1"))
    print(handler("foo.bar2"))
    print()

    handler = Handler(
        callback=[callback1, callback2], filter=filter, formatter=formatter
    )
    print("Test 4")
    print(handler("foo.bar1"))
    print(handler("foo.bar2"))
    print()

    def topic_callback1(topic, msg):
        print("callback1", topic, msg)
        return 1

    print("Test 5. Topic Handler")
    handler = TopicHandler()
    handler.add("a.b", callback=topic_callback1)
    handler("a", "")
    handler("b", "")
    handler("a.b", "")
    print()

    print("Test 6. Prefix Handler")
    handler = PrefixHandler()
    handler.add("a", callback=topic_callback1)
    handler("a", "")
    handler("b", "")
    handler("a.b", "")
    print()

    # aio stuff
    import asyncio

    async def aiocallback(topic="", msg=""):
        await asyncio.sleep(0.5)
        print("aiocallback", topic, msg)
        return 3

    loop = asyncio.get_event_loop()

    print("Test 7. AIO 1")
    handler = Handler(callback=[callback1, callback2, aiocallback], filter=filter)
    loop.run_until_complete(handler.call_async("foo.bar1"))
    loop.run_until_complete(handler.call_async("foo.bar2"))
    print()

    print("Test 7. AIO Topic Handler")
    handler = TopicHandler()
    handler.add("a.b", callback=aiocallback)
    loop.run_until_complete(handler.call_async("a", ""))
    loop.run_until_complete(handler.call_async("a", ""))
    loop.run_until_complete(handler.call_async("a.b", ""))

    print()

    print("Test 7. AIO Prefix Handler")
    handler = PrefixHandler()
    handler.add("a", callback=[topic_callback1, aiocallback])
    loop.run_until_complete(handler.call_async("a", ""))
    loop.run_until_complete(handler.call_async("b", ""))
    loop.run_until_complete(handler.call_async("a.b", ""))
    print()


if __name__ == "__main__":
    test()



import logging

logger = logging.getLogger(__name__)


class FiniteStateMachine:
    _states = ["blank"]
    _require_state_callbacks = False

    def __init__(self, state="blank"):
        self.state = state

    def set_state(self, state):
        if state not in self._states:
            raise Exception("Unknown status")

        if state != self.state:
            self.on_state_change(self.state, state)

        self.state = state

    def on_state_change(self, old_state, new_state):
        logger.debug(f"State change: {old_state} -> {new_state}")
        state_callback = getattr(self, f"on_{new_state}", None)
        if not state_callback:
            if self._require_state_callbacks:
                raise Exception(f"Callback for state {new_state} is not defined")
        else:
            state_callback()

    def get_state(self):
        raise NotImplementedError



import logging

import time
import multiprocessing
from requests_futures.sessions import FuturesSession
from utils_ak.networking.proxy import cast_requests_proxies
from utils_ak.callback_timer import CallbackTimer

logger = logging.getLogger(__name__)


class NoProxyAvailableException(Exception):
    pass


class AvailabilityProxyGuard:
    def __init__(
        self,
        proxy_config,
        check_proxy_availability=True,
        check_proxy_url=None,
        check_proxy_params=None,
        timeout=10,
        proxy_lives=3,
    ):
        # {proxy: max_usages_at_once}
        self.proxy_config = dict(proxy_config)

        # {proxy: available}
        self.counter = dict(proxy_config)
        self.max_proxy_lives = proxy_lives
        self.proxy_lives = {proxy: self.max_proxy_lives for proxy in proxy_config}

        # test proxies
        for proxy in self.proxy_config:
            if not self._is_proxy_valid(proxy):
                raise Exception(f"Proxy {proxy} is not valid")

        self.lock = multiprocessing.Lock()
        self.available_proxies = set(self.proxy_config.keys())

        self.timer = CallbackTimer(self.upd_availability, interval=300)
        self.check_proxy_availability = check_proxy_availability
        self.check_proxy_url = check_proxy_url
        self.check_proxy_params = check_proxy_params
        self.timeout = timeout

    def update_proxy_lives(self, proxy, success):
        if success:
            self.proxy_lives[proxy] = self.max_proxy_lives
        else:
            self.proxy_lives[proxy] -= 1

    @staticmethod
    def _is_proxy_valid(proxy):
        try:
            cast_requests_proxies(proxy)
        except Exception:
            return False
        else:
            return True

    @property
    def available_counter(self):
        return {
            proxy: available
            for proxy, available in self.counter.items()
            if proxy in self.available_proxies
        }

    @property
    def proxies(self):
        return list(self.counter.keys())

    @property
    def non_available_proxies(self):
        return set(self.proxies) - set(self.available_proxies)

    def on_acquire_proxy_callback(self, proxy):
        pass

    def acquire_proxy(self, timeout=None, sleep_timeout=0.01):
        started = time.time()

        while True:
            self.lock.acquire(True)

            self.timer.run_if_possible()

            if any(
                available > 0 for proxy, available in self.available_counter.items()
            ):
                # found some available proxy. Note: lock is still acquired
                break

            # no available proxy
            self.lock.release()

            if timeout and time.time() - started > timeout:
                raise NoProxyAvailableException("No proxy available at the moment")

            time.sleep(sleep_timeout)

        proxy = max(self.available_counter.items(), key=lambda x: x[1])[0]
        logger.debug(f"Acquiring proxy {proxy}")
        self.counter[proxy] -= 1
        self.on_acquire_proxy_callback(proxy)
        self.lock.release()
        return proxy

    def release_proxy(self, proxy, success=True):
        logger.debug(f"Releasing proxy {proxy}")
        with self.lock:
            self.counter[proxy] += 1
            self.update_proxy_lives(proxy, success)

            if (
                not success
                and self.proxy_lives[proxy] <= 0
                and proxy in self.available_proxies
            ):
                logger.debug(f"Proxy {proxy} is not available")
                self.available_proxies.remove(proxy)

    def upd_availability(self, failed_only=True):
        logger.debug(f"Start updating proxy availability. failed_only: {failed_only}")

        if not self.check_proxy_availability:
            return True

        if not self.check_proxy_url:
            raise Exception("Availability can be checked only with specified url")

        proxies = self.proxies if not failed_only else self.non_available_proxies

        if not proxies:
            logger.debug("No proxies to update")
            return

        session = FuturesSession(max_workers=300)

        responses = []
        for proxy in proxies:
            response = session.get(
                self.check_proxy_url,
                params=self.check_proxy_params,
                proxies=cast_requests_proxies(proxy),
                timeout=self.timeout,
            )
            responses.append((proxy, response))

        for proxy, response in responses:
            success = self._is_availability_response_ok(response)
            self.update_proxy_lives(proxy, success)

            if success:
                # succeed
                if proxy not in self.available_proxies:
                    logger.debug(f"Proxy {proxy} is available")
                    self.available_proxies.add(proxy)
            else:
                # fail
                if proxy in self.available_proxies:
                    logger.debug(f"Proxy {proxy} is not available")
                    self.available_proxies.remove(proxy)

        logger.debug("Stop updating proxy availability")

    @staticmethod
    def _is_availability_response_ok(response):
        try:
            res = response.result()
            assert res.status_code == 200
        except Exception:
            return False
        else:
            return True


if __name__ == "__main__":
    from utils_ak.log import configure_stream_logging

    configure_stream_logging(level=logging.DEBUG)
    import random

    proxy_guard = AvailabilityProxyGuard(
        {None: 5},
        check_proxy_url="https://api.binance.com/api/v1/depth",
        check_proxy_params={"symbol": "ETHBTC", "limit": 100},
    )
    proxy_guard.upd_availability(failed_only=False)

    for i in range(10):
        proxy_ = proxy_guard.acquire_proxy(timeout=None)

        print(proxy_, proxy_guard.counter)
        if random.randint(0, 2) == 0:
            proxy_guard.release_proxy(proxy_)



from datetime import datetime
from utils_ak.simple_microservice import SimpleMicroservice


class Monitor:
    def __init__(self, message_broker, heartbeat_timeout=10):
        self.microservice = SimpleMicroservice("Monitor", message_broker=message_broker)

        self.workers = {}  # {id: {status, state, last_heartbeat}}
        self.heartbeat_timeout = heartbeat_timeout

        self.microservice.add_timer(self._update_stalled, 3.0)
        self.microservice.add_callback("monitor_in", "heartbeat", self._on_heartbeat)
        self.microservice.add_callback("monitor_in", "state", self._on_state)
        self.microservice.register_publishers(["monitor_out"])

        # # todo: del, hardcode
        # self.received_messages_cache = []

    def _update_stalled(self):
        for worker_id, worker in self.workers.items():
            if "status" not in worker:
                # non-initialized
                continue

            if worker["status"] == "success":
                continue

            if (
                datetime.utcnow() - worker["last_heartbeat"]
            ).total_seconds() > self.heartbeat_timeout:
                self._update_status(worker_id, "stalled")

    def _on_heartbeat(self, topic, id):
        if id in self.workers:
            self.workers[id]["last_heartbeat"] = datetime.utcnow()

    def _update_status(self, worker_id, status):
        if status != self.workers[worker_id].get("status"):
            self.microservice.logger.info("Status updated", id=worker_id, status=status)
            self.microservice.publish(
                "monitor_out",
                "status_change",
                id=worker_id,
                old_status=self.workers[worker_id].get("status"),
                new_status=status,
                state=self.workers[worker_id].get("state", {}),
            )
            self.workers[worker_id]["status"] = status

    def _on_state(self, topic, id, status, state):
        # # todo: del, preview hardcode
        # if id is not None and state is not None:
        #     key = id + state
        #     if key not in self.received_messages_cache:
        #         self.received_messages_cache.append(key)
        #     else:
        #         return

        # init if necessary
        if id not in self.workers:
            self.microservice.publish("monitor_out", "new", id=id)
            self.workers[id] = {"last_heartbeat": datetime.utcnow()}

        self._update_status(id, status)

        # set state
        _old_state = self.workers[id].get("state")
        self.workers[id]["state"] = state
        if _old_state != state:
            self.microservice.logger.debug("New state", id=id, state=state)



from .broker import Broker

# from utils_ak.message_queue.clients.kafka_client import KafkaClient

import logging

logger = logging.getLogger(__name__)


class KafkaBroker(Broker):
    def __init__(self):
        self.cli = KafkaClient()
        self.async_supported = False  # todo: swithc to True?

    def _get_kafka_topic(self, collection, topic):
        if topic:
            return f"{collection}__{topic}"
        else:
            return collection

    def _split_kafka_topic(self, kafka_topic):
        if "__" in kafka_topic:
            return kafka_topic.split("__", 1)
        else:
            return kafka_topic, ""

    def publish(self, collection, topic, msg):
        self.cli.publish(self._get_kafka_topic(collection, topic), msg)

    def subscribe(self, collection, topic):
        self.cli.subscribe(self._get_kafka_topic(collection, topic))

    def poll(self, timeout=0.0):
        received_message = self.cli.poll(timeout)

        if not received_message:
            return

        kafka_topic, msg = received_message.topic(), received_message.value()

        if not kafka_topic:
            logging.error(
                f"Empty kafka_topic received: kafka_topic={kafka_topic}, msg={msg}"
            )
            raise Exception("Empty kafka_topic received")

        # todo: kafka_topic can be emtpy. Why?
        # print(received_message.offset())
        # msg["__kafka_offset"] = received_message.offset()

        collection, topic = self._split_kafka_topic(kafka_topic)
        return collection, topic, msg



import torch.nn as nn
import torch.utils.data
import numpy as np

from .images import *


class DynamicNet(nn.Module):
    """
    Pytorch net that constructs during first forward pass

    Sample Usage:

    class LeNet(DynamicNet):
        def _forward(self, x):
            x = self.layer('Conv2d', 1, 20, 5, 1)(x)
            x = F.relu(x)
            x = F.max_pool2d(x, 2, 2)

            x = self.layer('Conv2d', 20, 50, 5, 1)(x)
            x = F.relu(x)
            x = F.max_pool2d(x, 2, 2)

            x = x.view(-1, 4*4*50)
            x = self.layer('Linear', 4 * 4 * 50, 500)(x)
            x = F.relu(x)
            x = self.layer('Linear', 500, 10)(x)
            return x

    model = LeNet()
    # init net with one forward pass
    model.forward(Variable(X_test))

    # now model.parameters() is not empty
    """

    def __init__(self):
        super().__init__()
        self.init = True
        self.layer_counter = {}

    def forward(self, x):
        res = self._forward(x)
        self.init = False
        self.layer_counter = {}
        return res

    def layer(self, name, *args, **kwargs):
        if self.init:
            layer_count = 0
            while True:
                if not hasattr(self, f"{name}_{layer_count}"):
                    break
                layer_count += 1

            if hasattr(nn, name):
                res = getattr(nn, name)(*args, **kwargs)
            elif hasattr(self, name):
                res = getattr(self, name)(*args, **kwargs)
            else:
                raise Exception(f"Layer not found: {name}]")
            setattr(self, f"{name}_{layer_count}", res)
            return res
        else:
            layer_count = self.layer_counter.get(name, -1) + 1
            self.layer_counter[name] = layer_count
            return getattr(self, f"{name}_{layer_count}")

    def print_init(self, *args):
        if self.init:
            print(*args)


class NPYDataset(torch.utils.datasets.Dataset):
    def __init__(
        self,
        images_path,
        labels_path=None,
        transform=None,
        train=True,
        train_size=0.9,
        shuffle=False,
        seed=12,
    ):
        self.images = np.load(images_path)
        if labels_path:
            self.labels = np.load(labels_path)
        else:
            self.labels = None
        self.transform = transform

        self.total_size = self.images.shape[0]
        if isinstance(train_size, int):
            self.train_size = train_size
        else:
            self.train_size = int(self.total_size * train_size)
        self.train = train

        if not shuffle:
            train_idxs = np.arange(0, self.train_size)
            test_idxs = np.arange(self.train_size, self.total_size)
        else:
            np.random.seed(seed)
            permutation = np.random.permutation(self.total_size)
            train_idxs = permutation[: self.train_size]
            test_idxs = permutation[self.train_size :]

        if self.train:
            self.images = self.images[train_idxs, :, :]
            if self.labels is not None:
                self.labels = self.labels[train_idxs]
        else:
            self.images = self.images[test_idxs, :, :]
            if self.labels is not None:
                self.labels = self.labels[test_idxs]

    def __getitem__(self, index):
        im = self.images[index, :, :]

        if self.labels is not None:
            label = self.labels[index]
        else:
            label = 0

        if self.transform is not None:
            im = cast_pytorch_image(im)
            im = cast_image(im)
            im = self.transform(im)
        return im, label

    def __len__(self):
        return self.images.shape[0]


def collect_images_to_numpy(data_loader):
    data = None
    labels = None
    for cur_data, cur_labels in data_loader:
        data_numpy = (cur_data * 255).numpy().astype("uint8")
        labels_numpy = cur_labels.numpy()
        if data is None:
            data = data_numpy
            labels = labels_numpy
        else:
            data = np.concatenate([data, data_numpy])
            labels = np.concatenate([labels, labels_numpy])

    return data, labels


import logging



""" Miscellaneous functionality with basic python built-in objects. """
import collections
import numpy
import itertools


def delistify(lst_obj):
    if isinstance(lst_obj, list):
        return lst_obj if (len(lst_obj) > 1 or not lst_obj) else lst_obj[0]
    else:
        return lst_obj


def delistify_single(iterable):
    lst = list(iterable)
    assert len(lst) == 1
    return lst[0]


def cast_list(lst):
    return lst if isinstance(lst, list) else [lst]


listify = cast_list


def iter_get(iterable, index=0, default=None):
    iterator = iter(iterable)

    if index >= 0:
        try:
            return next(itertools.islice(iterator, index, None))
        except StopIteration:
            return default
    else:
        lst = list(iterator)
        try:
            return lst[index]
        except IndexError:
            return default


def iter_len(iterable):
    return len(list(iterable))


def remove_duplicates(seq, key=None):
    # preserves order of sequence
    seen = set()
    seen_add = seen.add
    key = key or (lambda x: x)
    return [x for x in seq if not (key(x) in seen or seen_add(key(x)))]


def remove_neighbor_duplicates(seq, key=None):
    key = key or (lambda x: x)
    res = []
    for x in seq:
        if res and key(res[-1]) == key(x):
            continue
        res.append(x)
    return res


def apply_on_slice(f, lst, cond):
    ind = [i for i, v in enumerate(lst) if cond(v)]
    applied = f([lst[i] for i in ind])
    res = list(lst)
    for j, i in enumerate(ind):
        res[i] = applied[j]
    return res


# NOTE: use anyconfig.merge for more advanced merge logics
def update_dic(dic, new_dic):
    """
    :param dic: obj, dict by default
    :param new_dic: dict
    :return:
    """
    for key, val in new_dic.items():
        if isinstance(dic, collections.Mapping):
            if isinstance(val, collections.Mapping):
                dic[key] = update_dic(dic.get(key, {}), val)
            else:
                dic[key] = new_dic[key]
        else:
            # dic is not a dictionary. Make it one
            dic = {key: new_dic[key]}
    return dic


def crop_to_chunks(lst, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def split_list(lst, n):
    """Split list to n parts with equal size."""
    for array in numpy.array_split(lst, n):
        yield list(array)


def unfold_dic(dic, keys, get=False, default=None):
    if get:
        return [dic.get(key, default) for key in keys]
    else:
        return [dic[key] for key in keys]


def recycle_list(lst, idx):
    return (lst * 2)[idx % len(lst) :][: len(lst)]


def test_collection():
    d1 = {"k1": {"k3": 4}}
    d2 = {"k1": {"k3": [2, 3]}}
    print(update_dic(d1, d2), d1, d2)

    print(
        apply_on_slice(
            lambda lst: [x ** 2 for x in lst], [1, 2, 3, 4, 5, 6], lambda x: x % 2 == 0
        )
    )
    # print(filter_dic({'k': 1, 'e': 2}, leave=2))
    print(
        remove_duplicates(
            [
                1,
                2,
                3,
                4,
                5,
                1,
                1,
            ]
        )
    )
    print(
        remove_neighbor_duplicates(
            [
                1,
                2,
                3,
                4,
                5,
                1,
                1,
            ]
        )
    )

    lst = [1, 2, 3, 4]
    print(lst)
    print(recycle_list(lst, 2))


if __name__ == "__main__":
    test_collection()



from utils_ak.str import trim


def get_source_code(method):
    import inspect

    lines = inspect.getsourcelines(method)[0]
    def_line = lines[0]
    offset_len = def_line.index(def_line.strip())
    offset = def_line[:offset_len]
    lines = [trim(l, beg=offset) for l in lines]
    return "".join(lines)



from functools import partial
from utils_ak.simple_vector import *
from utils_ak.properties import *
from utils_ak.block_tree import Block

from utils_ak.clock import *


def x_cumsum_acc(parent, child, key, default=None, formatter=None):
    if parent:
        return parent[key].add(
            relative_acc(parent, child, key, default=default, formatter=formatter)
        )
    return SimpleVector(
        list(
            relative_acc(
                parent, child, key, default=default, formatter=formatter
            ).values
        )
    )


class ParallelepipedBlock(Block):
    def __init__(self, block_class, n_dims=2, **props):
        self.n_dims = n_dims

        if "x" not in props:
            props["x"] = SimpleVector(n_dims)
        if "size" not in props:
            props["size"] = SimpleVector(n_dims)

        super().__init__(
            block_class,
            props_formatters={
                "x": lambda k, v: SimpleVector(v),
                "size": lambda k, v: SimpleVector(v),
            },
            **props,
        )
        self.props.accumulators["x"] = x_cumsum_acc
        self.props.accumulators[
            "x_rel"
        ] = lambda parent_props, child_props, key: relative_acc(
            parent_props, child_props, "x"
        )
        self.props.accumulators["size"] = relative_acc
        self.props.accumulators["axis"] = partial(relative_acc, default=0)
        self.props.accumulators["is_parent_node"] = partial(relative_acc, default=False)

    @property
    def x(self):
        return self.props["x"]

    @property
    def y(self):
        return self.x + self.size

    def __str__(self):
        res = (
            self.props["cls"]
            + " "
            + " x ".join([f"({self.x[i]}, {self.y[i]}]" for i in range(self.n_dims)])
        )

        for child in self.children:
            for line in str(child).split("\n"):
                if not line:
                    continue
                res += "\n  " + line
        return res

    def __repr__(self):
        return self.tabular()

    def is_leaf(self):
        return not self.children and not self.props["is_parent_node"]

    def tabular(self):
        res = ""
        for b in self.iter():
            if b.size[0] != 0:
                res += (
                    " " * int(b.x[0])
                    + "=" * int(b.size[0])
                    + f' {b.props["cls"]} '
                    + " x ".join([f"({b.x[i]}, {b.y[i]}]" for i in range(b.n_dims)])
                )
                res += "\n"
        return res

    @property
    def size(self):
        size = self.props["size"]
        values = []
        for axis in range(self.n_dims):
            if size[axis] == 0:
                if not self.children:
                    values.append(0)
                else:
                    start = min([c.x[axis] for c in self.children] + [self.x[axis]])
                    # start = self.x[axis]
                    values.append(max([c.y[axis] - start for c in self.children]))
            else:
                values.append(size[axis])
        return SimpleVector(values)


def test_parallelepiped_block():
    a = ParallelepipedBlock("a", n_dims=2, x=[1, 2])
    b = ParallelepipedBlock("b", n_dims=2)
    c = ParallelepipedBlock("c", n_dims=2, x=[3, 4], size=[1, 5])
    a.add_child(b)
    b.add_child(c)

    print(a)
    print(b)
    print(c)
    print(a.x, a.size, a.y)
    print(b.x, b.size, b.y)
    print(c.x, c.size, c.y)

    print()
    print(a["b"]["c"])

    print(a.__repr__())


if __name__ == "__main__":
    test_parallelepiped_block()



# from confluent_kafka import Producer, Consumer
import uuid

from utils_ak.builtin import update_dic
from copy import deepcopy

# todo: make group_id properly

DEFAULT_CONSUMER_CONFIG = {
    "bootstrap.servers": "localhost:9092",
    "group.id": str(uuid.uuid4()),
    "default.topic.config": {"auto.offset.reset": "largest"},
    "enable.auto.commit": False,
    "enable.partition.eof": False,
}

DEFAULT_PRODUCER_CONFIG = {
    "bootstrap.servers": "localhost:9092",
    "queue.buffering.max.ms": 1,
    "queue.buffering.max.messages": 1000000,
    "max.in.flight.requests.per.connection": 1,
    "default.topic.config": {"acks": "all"},
}


class KafkaClient:
    def __init__(self, consumer_config=None, producer_config=None):
        self.kafka_topics = []

        consumer_config = consumer_config or {}
        self.consumer_config = deepcopy(DEFAULT_CONSUMER_CONFIG)
        self.consumer_config = update_dic(self.consumer_config, consumer_config)
        self.consumer = Consumer(self.consumer_config)

        producer_config = producer_config or {}
        self.producer_config = deepcopy(DEFAULT_PRODUCER_CONFIG)
        self.producer_config = update_dic(self.producer_config, producer_config)
        self.producer = Producer(self.producer_config)

        self.init_subscriptions = False

    def subscribe(self, topic):
        if topic not in self.kafka_topics:
            self.kafka_topics.append(topic)
            # self.consumer.subscribe(self.kafka_topics)

    def publish(self, topic, msg):
        self.producer.produce(topic, msg)
        self.producer.poll(0)

    def poll(self, timeout=0.0):
        self.start_listening()
        return self.consumer.poll(timeout)

    def start_listening(self):
        if not self.init_subscriptions:
            self.consumer.subscribe(self.kafka_topics)
            self.init_subscriptions = True


if __name__ == "__main__":
    cli = KafkaClient(
        consumer_config={"default.topic.config": {"auto.offset.reset": "smallest"}}
    )
    cli.subscribe("binance_trades")

    i = 0
    while True:
        msg = cli.poll()
        if not msg:
            continue
        print(msg, msg.offset(), msg.value())
        print(i)

        i += 1



from bson.objectid import ObjectId
from mongoengine import Document


def cast_object_id(obj):
    if isinstance(obj, ObjectId):
        return obj
    elif isinstance(obj, str):
        return ObjectId(obj)
    elif obj is None:
        return None
    elif isinstance(obj, dict):
        return obj["_id"]
    else:
        raise Exception("Unknown object id type")


def cast_model_object(obj, cls):
    if isinstance(obj, ObjectId) or isinstance(obj, str):
        object_id = cast_object_id(obj)
        return cast_model_object({"_id": object_id}, cls)
    elif isinstance(obj, dict):
        if "_id" in obj:
            # fetch and return updated version from server
            element = cls.objects(pk=obj["_id"]).first()
            if not element:
                raise Exception("Object not found")
            return element
        else:
            # init
            return cls(**obj)
    elif isinstance(obj, cls):
        return obj
    else:
        raise Exception("Unknown model format")


def cast_dict(obj):
    if isinstance(obj, dict):
        return obj
    elif isinstance(obj, Document):
        res = dict(obj.to_mongo())
        res["_cls"] = obj.__class__.__name__
        return res
    else:
        raise Exception("Unknown dict type")



class StateProvider:
    def get_state(self):
        pass

    def set_state(self, state):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import pandas as pd
import fnmatch
import os

from utils_ak.os import *

from loguru import logger


class ScreenClient:
    def start(self, name, cmd):
        makedirs("screen_logs/")
        logger.info(f"Starting screen {name} with command {cmd}")
        execute(
            f'screen -dm -S {name} -L -Logfile screen_logs/{name}.log bash -c "{cmd}"',
            is_async=True,
        )

    def start_python(self, name, path, **kwargs):
        if not os.path.exists(path):
            raise Exception(f"File does not exist {path}")

        cmd = f"python {path}"
        for k, v in kwargs.items():
            cmd += f"--{k} {v} "

        self.start(name, cmd)

    def _filter(self, df, filter):
        if isinstance(filter, str):
            _filter = str(filter)
            filter = lambda s: fnmatch.fnmatch(s, _filter)
        if filter:
            df = df[df["name"].apply(filter)]
        return df

    def list(self, filter=None, order_by="time"):
        output = execute("screen -ls", is_async=False)
        lines = output.split("\n")[1:-2]

        values = []
        for line in lines:
            value = line.split("\t")[1:]
            value = value[0].split(".", 1) + value[1:]
            if len(value) == 3:
                # no time provided
                value = value[:2] + [""] + value[2:3]
            values.append(value)

        df = pd.DataFrame(values, columns=["id", "name", "time", "status"])
        df["status"] = df["status"].apply(lambda s: s.replace("(", "").replace(")", ""))
        df["time"] = df["time"].apply(lambda s: s.replace("(", "").replace(")", ""))
        df["time"] = pd.to_datetime(df["time"])
        df = df[["time", "name", "id", "status"]]
        df = df.sort_values(by=order_by)
        df = self._filter(df, filter)
        df = df.reset_index(drop=True)
        return df

    def kill_one(self, name):
        logger.info(f"Killing screen {name}")
        return execute(f"screen -X -S {name} quit", is_async=False)

    def kill(self, filter):
        df = self.list()

        if len(df) == 0:
            return

        df["full_name"] = df["id"].astype(str) + "." + df["name"]
        df = self._filter(df, filter)

        for i, row in df.iterrows():
            self.kill_one(row["full_name"])

    def restart(self, name, cmd):
        self.kill(name)
        self.start(name, cmd)


def test():
    import time

    cli = ScreenClient()
    cli.start("test", "sleep 10")
    print(cli.list())
    print(cli.list("t*"))
    print(cli.list("b*"))  # empty
    time.sleep(1)
    cli.kill("t*")
    print(cli.list())

    print(cli.start_python("hello-world", "tests/hello_world.py"))
    print(cli.list("hello-world"))
    time.sleep(1)
    cli.kill("hello-world")
    print(cli.list("hello-world"))


if __name__ == "__main__":
    test()



import calendar
from datetime import datetime, timedelta


def add_months(dt, n):
    month = dt.month - 1 + n
    year = dt.year + month // 12
    month = month % 12 + 1
    day = min(dt.day, calendar.monthrange(year, month)[1])
    return datetime(year, month, day)


def iter_range_by_months(beg, end):
    while True:
        next_end = add_months(
            datetime(beg.year, beg.month, 1), 1
        )  # first day of next month
        next_end = min(next_end, end)
        yield beg, next_end
        beg = next_end

        if next_end == end:
            return


def _get_quarter_by_month(month_name):
    if month_name in range(1, 4):
        return "Q1"
    elif month_name in range(4, 7):
        return "Q2"
    elif month_name in range(7, 10):
        return "Q3"
    elif month_name in range(10, 13):
        return "Q4"


def iter_quarters(beg, end):
    cur_beg = beg
    for _beg, _end in iter_range_by_months(beg, end):
        if _end.month in [4, 7, 10, 1]:
            yield _beg.year, _get_quarter_by_month(_beg.month), cur_beg, _end
            cur_beg = _end

        if _end == end and cur_beg != _end:
            # last value
            yield _beg.year, _get_quarter_by_month(_beg.month), cur_beg, _end


def iter_range(beg, end, period):
    while beg < end:
        next_beg = beg + period
        next_beg = min(next_beg, end)
        yield beg, next_beg
        beg = next_beg


def datetime_range(beg, period=None, end=None, n=None):
    if end:
        while beg < end:
            yield beg
            beg = min(beg + period, end)
    elif n:
        for i in range(n):
            yield beg + i * period
    else:
        raise Exception("end or n not specified")


def test_ranges():
    from utils_ak.time import cast_datetime, cast_timedelta

    # add months
    dt = cast_datetime("2020.01.30")
    for i in range(12):
        print(add_months(dt, i))

    # iter range by months
    print(
        list(
            iter_range_by_months(
                cast_datetime("2020.01.15"), cast_datetime("2020.02.16")
            )
        )
    )
    print(
        list(
            iter_range_by_months(
                cast_datetime("2020.01.01"), cast_datetime("2020.06.01")
            )
        )
    )

    # iter range
    print(
        list(
            iter_range(
                cast_datetime("2020.01.01 12:00:00"),
                cast_datetime("2020.01.01 12:16:00"),
                cast_timedelta("5m"),
            )
        )
    )

    print(list(iter_quarters(cast_datetime("2020.02.15"), cast_datetime("2020.07.15"))))

    print(
        list(
            datetime_range(
                cast_datetime("2020.01.01"),
                period=timedelta(days=1),
                end=cast_datetime("2020.01.04"),
            )
        )
    )
    print(
        list(datetime_range(cast_datetime("2020.01.01"), period=timedelta(days=1), n=3))
    )


if __name__ == "__main__":
    test_ranges()



import asyncio
from datetime import datetime

from utils_ak.job_orchestrator.worker.worker import MicroserviceWorker
from utils_ak.job_orchestrator.worker.test_worker import *


class TestWorker(MicroserviceWorker):
    async def process(self):
        self.started = datetime.now()

        running_timeout = self.payload.get("running_timeout", 0)
        if self.payload.get("type") == "batch":
            time.sleep(1)
            time.sleep(self.payload.get("initializing_timeout", 0))
            self.send_state("running", {})
            time.sleep(self.payload.get("stalled_timeout", 0))
            N = 10
            for i in range(N):
                self.send_state("running", {"progress": (i + 1) * 100 / N})
                await asyncio.sleep(0.1 + running_timeout / N)
            self.send_state("success", {"response": "42"})
        elif self.payload.get("type") == "streaming":
            time.sleep(3)
            time.sleep(self.payload.get("initializing_timeout", 0))
            self.send_state("running", {})
            time.sleep(self.payload.get("running_timeout", 0))
            while True:
                self.send_state("running", {"foo": "bar"})
                time.sleep(3)
                if (datetime.now() - self.started).total_seconds() > 10:
                    # work for one-minute only and than die
                    break
        else:
            raise Exception(f'Bad payload type {self.payload.get("type")}')

        self.microservice.stop()



import anyconfig
from utils_ak.coder import *
from utils_ak.builtin import *
import fire

# todo: put to deployment util


def generate_bats_from_deployment(deployment_fn="sample_deployment.yml"):
    d = anyconfig.load(deployment_fn)

    assert (
        len(list(d["containers"])) == 1
    ), "Only single container allowed at the moment"

    container = iter_get(d["containers"].values())
    config_js = cast_js(container["command_line_arguments"]["config"])
    config_js = config_js.replace('"', r"\"")
    config_js = f'"{config_js}"'

    image = container["image"]

    with open("build.bat", "w") as f:
        f.write(f"docker build -t {image}:latest . --no-cache\n")
        f.write("pause")

    with open("push.bat", "w") as f:
        f.write(f"docker push {image}:latest\n")
        f.write("pause")

    with open("run.bat", "w") as f:
        f.write(f"""docker run {image} --config {config_js}\n""")
        f.write("pause")


if __name__ == "__main__":
    fire.Fire(generate_bats_from_deployment)



""" TopicHandler is a builtin of handlers, bound to certain topics as callbacks. """
from .handler import Handler
from utils_ak.builtin import delistify
from sortedcollections import OrderedDict

# todo: make handlers a list for better priorities?


class TopicHandler(object):
    """ Run handlers on events. Each event has topic as it's id, which is passed as argument to handler functions. """

    def __init__(self, topic_formatter=None, topic_filter=None, reducer=None):
        self.handlers = OrderedDict()  # {topic: handler}
        self.topic_formatter = topic_formatter
        self.topic_filter = topic_filter or (
            lambda topic, received_ropic: topic == received_ropic
        )
        self.reducer = reducer or delistify

    def add(self, topic, callback=None, formatter=None, filter=None, reducer=None):
        handler = Handler(reducer=reducer)
        handler.add(callback=callback, filter=filter, formatter=formatter)
        self.handlers.setdefault(topic, Handler()).add(callback=handler)

    def has_coroutine(self, topic):
        return self.handlers[topic].has_coroutine()

    def set_topic_formatter(self, formatter):
        self.topic_formatter = formatter

    def __call__(self, topic, *args, **kwargs):
        args = [topic] + list(args)
        if self.topic_formatter:
            topic = self.topic_formatter(topic)

        res = []
        for _topic, handler in self.handlers.items():
            if self.topic_filter(_topic, topic):
                res.append(handler(*args, **kwargs))
        return self.reducer(res)

    def call(self, topic, *args, **kwargs):
        return self.__call__(topic, *args, **kwargs)

    async def call_async(self, topic, *args, **kwargs):
        args = [topic] + list(args)
        if self.topic_formatter:
            topic = self.topic_formatter(topic)

        res = []
        for _topic, handler in self.handlers.items():
            if self.topic_filter(_topic, topic):
                res.append(await handler.call_async(*args, **kwargs))
        return self.reducer(res)

    def has_topic(self, topic):
        return topic in self.handlers

    def __getitem__(self, item):
        return self.handlers[item]


class PrefixHandler(TopicHandler):
    def __init__(self, topic_formatter=None, reducer=None):
        super().__init__(
            topic_formatter=topic_formatter,
            topic_filter=lambda topic, received_topic: received_topic.startswith(topic),
            reducer=reducer,
        )



from datetime import datetime

from utils_ak.loguru import *
from utils_ak.deployment import *
from utils_ak.simple_microservice import SimpleMicroservice
from utils_ak.dict import fill_template
from utils_ak.coder.coders.json import cast_dict_or_list

from .models import Job, Worker


# todo: retry


class JobOrchestrator:
    def __init__(self, deployment_controller, message_broker):
        self.controller = deployment_controller
        self.microservice = SimpleMicroservice(
            "JobOrchestrator", message_broker=message_broker
        )
        self.microservice.add_timer(self._process_pending, 1.0)
        self.microservice.add_timer(self._process_initializing, 5.0)
        self.microservice.add_timer(self._process_running, 5.0)
        self.microservice.add_callback("monitor_out", "status_change", self._on_monitor)
        self.microservice.register_publishers(["job_orchestrator"])

    def run(self):
        self.microservice.run()

    def _update_job_status(self, worker, status, response):
        logger.debug(
            "Updating job status", id=worker.job.id, status=status, response=response
        )

        if status == "success":
            logger.info(
                "Successfully processed job", id=worker.job.id, response=response
            )
        elif status == "error":
            logger.error("Failed to process job", id=worker.job.id, response=response)

        self.microservice.publish(
            "job_orchestrator",
            "status_change",
            id=str(worker.job.id),
            old_status=worker.job.status,
            new_status=status,
            response=response,
        )
        worker.response = response
        worker.job.status = status
        worker.job.save()

    def _update_worker_status(self, worker, status, state=None):
        logger.debug("Updating worker status", id=worker.id, status=status, state=state)
        state = state or {}
        self._update_job_status(worker, status, state.get("response", ""))

        worker.status = status
        worker.save()

        if worker.status in ["success", "stalled", "error"]:
            self._stop_worker(worker)

    def _stop_worker(self, worker):
        logger.info("Stopping worker", id=worker.id)
        self.controller.stop(str(worker.id))

    def _create_worker_model(self, job):
        worker_model = Worker()
        worker_model.job = job
        worker_model.save()
        job.workers.append(worker_model)
        job.locked_by = worker_model
        job.locked_at = datetime.utcnow()
        job.save()
        return worker_model

    def _create_deployment(self, worker_model):
        deployment = cast_dict_or_list(
            os.path.join(os.path.dirname(__file__), "worker/deployment.yml.template")
        )

        params = {
            "deployment_id": str(worker_model.id),
            "payload": worker_model.job.payload,
            "image": worker_model.job.runnable.get("image", ""),
            "python_main": worker_model.job.runnable.get("python_main", ""),
        }
        deployment = fill_template(deployment, **params)
        return deployment

    def _process_pending(self):
        jobs = Job.objects(status="pending").all()

        if jobs:
            self.microservice.logger.info("Processing: pending", n=len(jobs))

        for job in jobs:
            worker_model = self._create_worker_model(job)
            self._update_worker_status(worker_model, "initializing")
            deployment = self._create_deployment(worker_model)
            self.controller.start(deployment)

    def _process_initializing(self):
        jobs = Job.objects(status="initializing").all()

        if jobs:
            self.microservice.logger.info("Processing: initializing", n=len(jobs))

        for job in jobs:
            worker = job.locked_by
            assert worker is not None, "Worker not assigned for the job"
            if (
                job.initializing_timeout
                and (datetime.utcnow() - worker.job.locked_at).total_seconds()
                > job.initializing_timeout
            ):
                logger.error(
                    "Initializing timeout expired",
                    worker_id=worker.id,
                    locked_at=worker.job.locked_at,
                )
                self._update_worker_status(
                    worker,
                    "error",
                    {"response": "Initializing timeout expired"},
                )

    def _process_running(self):
        jobs = Job.objects(status="running").all()

        if jobs:
            self.microservice.logger.info("Processing: running", n=len(jobs))

        for job in jobs:
            worker = job.locked_by
            assert worker is not None, "Worker not assigned for the job"
            if (
                job.running_timeout
                and (datetime.utcnow() - worker.job.locked_at).total_seconds()
                > job.running_timeout
            ):
                logger.error(
                    "Running timeout expired",
                    worker_id=worker.id,
                    locked_at=worker.job.locked_at,
                )
                self._update_worker_status(
                    worker,
                    "error",
                    {"response": "Running timeout expired"},
                )

    def _on_monitor(self, topic, id, old_status, new_status, state):
        try:
            worker = Worker.objects(pk=id).first()  # todo: check if missing
        except:
            logger.error("Failed to fetch worker", id=id)
            return

        # check if locked by another worker
        if str(id) != str(worker.job.locked_by.id):
            logger.error(
                "Job is locked by another worker",
                worker_id=id,
                another_worker_id=worker.job.locked_by.id,
            )
            self._stop_worker(worker)
            return

        status = new_status
        if status in ["stalled", "error"]:
            # todo: make properly
            if "response" not in state:
                state["response"] = status
            status = "error"
        self._update_worker_status(worker, status, state)



from queue import Queue
import asyncio
import concurrent.futures

from utils_ak.architecture.finite_state_machine import FiniteStateMachine
from utils_ak.architecture.sink import ReturnValuesSink

import logging

logger = logging.getLogger(__name__)


class TaskOperation(FiniteStateMachine):
    _states = ["blank", "initiated", "started", "failed", "finished"]

    def __init__(self):
        super().__init__()
        # task is anything
        self.tasks = []

    def on_initiated(self):
        pass

    def on_failed(self):
        pass

    def on_finished(self):
        pass

    def on_started(self):
        pass

    def check_state(self):
        raise NotImplementedError

    def initiate(self, tasks):
        self.tasks = tasks
        self.set_state("initiated")

    def start(self):
        self.set_state("started")

    def finish(self):
        self.set_state("finished")

    def fail(self):
        self.set_state("failed")


class AsyncioTaskOperation(TaskOperation):
    def __init__(
        self,
        results_sink=None,
        integrity_guard=None,
        max_workers=20,
        operation_name=None,
        max_retries=10,
    ):
        super().__init__()
        self.fired_tasks_queue = Queue()
        self.tasks_queue = Queue()

        self.integrity_guard = integrity_guard
        self.results_sink = results_sink or ReturnValuesSink()

        self.max_workers = max_workers
        self.operation_name = operation_name

        self.fail_alert = False

        self.max_retries = max_retries

        # {task_name: number_of_retries}
        self.retry_counter = {}

    def initiate(self, tasks):
        self.tasks = tasks
        for task in tasks:
            self.tasks_queue.put(task)
            self.retry_counter[task.name] = 0
        self.set_state("initiated")

    def check_integrity(self):
        if not self.integrity_guard:
            return True

        if not self.integrity_guard.is_integral():
            self.set_state("failed")
            return False
        return True

    def on_state_change(self, old_state, new_state):
        super().on_state_change(old_state, new_state)

        if self.integrity_guard:
            if new_state in ["failed", "finished"]:
                self.integrity_guard.leave_unstable_state()
            elif new_state in ["started"]:
                self.integrity_guard.enter_unstable_state()

    async def execute(self):
        if not self.check_integrity():
            return

        self.set_state("started")

        executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers)

        loop = asyncio.get_event_loop()
        loop.set_default_executor(executor)

        while not self.tasks_queue.empty():
            while not self.tasks_queue.empty():
                if self.fail_alert:
                    break

                task = self.tasks_queue.get()

                self.fired_tasks_queue.put(
                    loop.run_in_executor(executor, self._execute_task, task)
                )

                self.tasks_queue.task_done()

            if self.fail_alert:
                break

            while not self.fired_tasks_queue.empty():
                await self.fired_tasks_queue.get()

        self.set_state("failed" if self.fail_alert else "finished")

    def _execute_task(self, task):
        try:
            res = task()
        except Exception as e:
            self.retry_counter[task.name] += 1

            if self.retry_counter[task.name] >= self.max_retries:
                self.fail_alert = True
            else:
                self.tasks_queue.put(task)
            logger.warning(f"Failed to process task, error: {e}")
        else:
            if self.results_sink:
                self.results_sink.store(task.name, res)

    def get_results(self):
        return self.results_sink.get_results()


if __name__ == "__main__":
    import time
    from utils_ak.log import configure_stream_logging

    configure_stream_logging(level=logging.DEBUG)

    from utils_ak.architecture.integrity_guard import FilesystemIntegrityGuard

    def gen_runnable(name):
        class Task:
            def __init__(self):
                self.name = name

            def __call__(self, *args, **kwargs):
                logger.info(f"Runnable {name} started...")
                time.sleep(1)
                if name == 4:
                    raise Exception("Fail example")
                logger.info(f"Runnable {name} finished...")

        return Task()

    operation = AsyncioTaskOperation(integrity_guard=FilesystemIntegrityGuard())
    operation.initiate([gen_runnable(i) for i in range(5)])
    asyncio.get_event_loop().run_until_complete(operation.execute())
